import unittest

from rlcard.games.doudizhu.utils import CARD_TYPE
from rlcard.games.doudizhu.judger import DoudizhuJudger as Judger

class TestDoudizhuGame(unittest.TestCase):

    def test_playable_cards_from_hand(self):
        # #solo
        # in_cards, not_in_cards, hand = ('3', '4', '5', '6', '8', '9', 'J', 'Q', 'B', 'R'), (), '334445555689JQBR'
        # playable_cards = Judger.playable_cards_from_hand(hand)
        # for e in in_cards:
        #     self.assertIn(e, playable_cards)
        # for e in not_in_cards:
        #     self.assertNotIn(e, playable_cards)

        # #pair
        # in_cards, not_in_cards, hand = ('33', '44', '55'), (), '334445555689JQBR'
        # playable_cards = Judger.playable_cards_from_hand(hand)
        # for e in in_cards:
        #     self.assertIn(e, playable_cards)
        # for e in not_in_cards:
        #     self.assertNotIn(e, playable_cards)
 
        # #trio
        # in_cards, not_in_cards, hand = ('444', '555'), (), '334445555689JQBR'
        # playable_cards = Judger.playable_cards_from_hand(hand)
        # for e in in_cards:
        #     self.assertIn(e, playable_cards)
        # for e in not_in_cards:
        #     self.assertNotIn(e, playable_cards)

        # #bomb
        # in_cards, not_in_cards, hand = ('5555', ), (), '334445555689JQBR'
        # playable_cards = Judger.playable_cards_from_hand(hand)
        # for e in in_cards:
        #     self.assertIn(e, playable_cards)
        # for e in not_in_cards:
        #     self.assertNotIn(e, playable_cards)

        # #rocket
        # in_cards, not_in_cards, hand = ('BR', ), (), '334445555689JQBR'
        # playable_cards = Judger.playable_cards_from_hand(hand)
        # for e in in_cards:
        #     self.assertIn(e, playable_cards)
        # for e in not_in_cards:
        #     self.assertNotIn(e, playable_cards)

        # #solo_chain_5 -- #solo_chain_12
        # in_cards = ('34567', '345678', '3456789', '3456789T', '3456789TJ',
        #     '3456789TJQ', '3456789TJQK', '3456789TJQKA', '45678', '456789',
        #     '3456789T', '3456789TJ', '456789TJQ', '456789TJQK', '456789TJQKA',
        #     '56789', '56789T', '56789TJ', '56789TJQ', '56789TJQK', '56789TJQKA',
        #     '6789T', '6789TJ', '6789TJQ', '6789TJQK', '6789TJQKA', '789TJ',
        #     '789TJQ', '789TJQK', '789TJQKA', '89TJQ', '89TJQK',
        #     '89TJQKA', '9TJQK', '9TJQKA', 'TJQKA')
        # not_in_cards = ('JQKA2', )
        # hand = '3344455556789TJQKA2BR'
        # playable_cards = Judger.playable_cards_from_hand(hand)
        # for e in in_cards:
        #     self.assertIn(e, playable_cards)
        # for e in not_in_cards:
        #     self.assertNotIn(e, playable_cards)

        # #pair_chain_3 -- #pair_chain_10
        # in_cards = ('334455', '33445566', '3344556677', '334455667788', '33445566778899',
        # '33445566778899TT', '33445566778899TTJJ', '33445566778899TTJJQQ', '445566', '44556677',
        # '4455667788', '445566778899', '445566778899TT', '445566778899TTJJ', '445566778899TTJJQQ',
        # '445566778899TTJJQQKK', '556677', '55667788', '5566778899', '5566778899TT',
        # '5566778899TTJJ', '5566778899TTJJQQ', '5566778899TTJJQQKK', '5566778899TTJJQQKKAA',
        # '667788', '66778899', '66778899TT', '66778899TTJJ', '66778899TTJJQQ',
        # '66778899TTJJQQKK', '66778899TTJJQQKKAA', '778899', '778899TT', '778899TTJJ',
        # '778899TTJJQQ', '778899TTJJQQKK', '778899TTJJQQKKAA', '8899TT', '8899TTJJ',
        # '8899TTJJQQ', '8899TTJJQQKK', '8899TTJJQQKKAA', '99TTJJ', '99TTJJQQ',
        # '99TTJJQQKK', '99TTJJQQKKAA', 'TTJJQQ', 'TTJJQQKK', 'TTJJQQKKAA',
        # 'JJQQKK', 'JJQQKKAA', 'QQKKAA')
        # not_in_cards = ('33445566778899TTJJQQKK', 'KKAA22')
        # hand =  '33444555566778899TTJJQQKKAA22'
        # playable_cards = Judger.playable_cards_from_hand(hand)
        # for e in in_cards:
        #     self.assertIn(e, playable_cards)
        # for e in not_in_cards:
        #     self.assertNotIn(e, playable_cards)

        # #trio_chain_2 -- #trio_chain_6
        # in_cards = ('333444', '333444555', '333444555666', '333444555666777', '333444555666777888',
        # '444555', '444555666', '444555666777', '444555666777888', '444555666777888999',
        # '555666', '555666777', '555666777888', '555666777888999', '555666777888999TTT',
        # '666777', '666777888', '666777888999', '666777888999TTT', '666777888999TTTJJJ',
        # '777888', '777888999', '777888999TTT', '777888999TTTJJJ', '777888999TTTJJJQQQ',
        # '888999', '888999TTT', '888999TTTJJJ', '888999TTTJJJQQQ', '888999TTTJJJQQQKKK',
        # '999TTT', '999TTTJJJ', '999TTTJJJQQQ', '999TTTJJJQQQKKK', '999TTTJJJQQQKKKAAA',
        # 'TTTJJJ', 'TTTJJJQQQ', 'TTTJJJQQQKKK', 'TTTJJJQQQKKKAAA',
        # 'JJJQQQ', 'JJJQQQKKK', 'JJJQQQKKKAAA',
        # 'QQQKKK', 'QQQKKKAAA',
        # 'KKKAAA')
        # not_in_cards = ('333444555666777888999', 'AAA222')
        # hand =  '333444455556667778889999TTTJJJQQQKKKAAA222BR'
        # playable_cards = Judger.playable_cards_from_hand(hand)
        # for e in in_cards:
        #     self.assertIn(e, playable_cards)
        # for e in not_in_cards:
        #     self.assertNotIn(e, playable_cards)

        # #trio_solo, trio_pair
        # in_cards = ('3777', '4777', '5777', '6777', '7778',
        # '7772', '44777', '55777', '77788', '77722')
        # not_in_cards = ()
        # hand =  '344455677778889TJQKA2222BR'
        # playable_cards = Judger.playable_cards_from_hand(hand)
        # for e in in_cards:
        #     self.assertIn(e, playable_cards)
        # for e in not_in_cards:
        #     self.assertNotIn(e, playable_cards)

        # #trio_solo_chain_2 -- trio_solo_chain_5
        # in_cards = ('34777888', '3777888T', '3456777888999TTTJJJQ', '66777888')
        # not_in_cards = ('37777888', '777888999TTTJJJQK2BR', '777888999TTTJJJJ')
        # hand =  '34556677778888999TTTTJJJJQQQQKA2BR'
        # playable_cards = Judger.playable_cards_from_hand(hand)
        # for e in in_cards:
        #     self.assertIn(e, playable_cards)
        # for e in not_in_cards:
        #     self.assertNotIn(e, playable_cards)

        # #trio_pair_chain_2 -- #trio_pair_chain_4
        # in_cards = ('5566777888', '55777888TT')
        # not_in_cards = ('777888QQQQ', )
        # hand =  '34556677778888999TTTTJJJJQQQQKA2BR'
        # playable_cards = Judger.playable_cards_from_hand(hand)
        # for e in in_cards:
        #     self.assertIn(e, playable_cards)
        # for e in not_in_cards:
        #     self.assertNotIn(e, playable_cards)

        # #four_two_solo, #four_two_pair
        # in_cards = ('357777', '557777', '567777', '777788', '55667777', '66777799', '557777TT', '55777788')
        # not_in_cards = ('77778888', )
        # hand =  '34556677778888999TTTTJJJJQQQQKA2BR'
        # playable_cards = Judger.playable_cards_from_hand(hand)
        # for e in in_cards:
        #     self.assertIn(e, playable_cards)
        # for e in not_in_cards:
        #     self.assertNotIn(e, playable_cards)

        playable_cards = list(Judger.playable_cards_from_hand('3333444455556666777788889999TTTTJJJJQQQQKKKKAAAA2222BR'))
        all_cards_list = CARD_TYPE[1]
        for c in playable_cards:
            # if (c not in all_cards_list):
            #     print(c)
            self.assertIn(c, all_cards_list)
        for c in all_cards_list:
            # if (c not in playable_cards):
            #     print('\t' + c)
            self.assertIn(c, playable_cards)
        self.assertEqual(len(playable_cards), len(all_cards_list))

if __name__ == '__main__':
    unittest.main()



import unittest

import rlcard
from rlcard.agents.random_agent import RandomAgent
from rlcard.games.nolimitholdem.round import Action
from .determism_util import is_deterministic


class TestNolimitholdemEnv(unittest.TestCase):

    def test_reset_and_extract_state(self):
        env = rlcard.make('no-limit-holdem')
        state, _ = env.reset()
        self.assertEqual(state['obs'].size, 54)

    def test_is_deterministic(self):
        self.assertTrue(is_deterministic('no-limit-holdem'))

    def test_get_legal_actions(self):
        env = rlcard.make('no-limit-holdem')
        env.reset()
        legal_actions = env._get_legal_actions()
        for action in legal_actions:
            self.assertIn(action, env.actions)

    def test_decode_action(self):
        env = rlcard.make('no-limit-holdem')
        state, _ = env.reset()
        for action in state['legal_actions']:
            decoded = env._decode_action(action)
            self.assertIn(decoded, env.actions)

        decoded = env._decode_action(Action.FOLD.value)
        self.assertEqual(decoded, Action.FOLD)

        env.step(0)
        decoded = env._decode_action(1)
        self.assertEqual(decoded, Action.CHECK_CALL)

    def test_step(self):
        env = rlcard.make('no-limit-holdem')
        state, player_id = env.reset()
        self.assertEqual(player_id, env.get_player_id())
        action = list(state['legal_actions'].keys())[0]
        _, player_id = env.step(action)
        self.assertEqual(player_id, env.get_player_id())

    def test_step_back(self):
        env = rlcard.make('no-limit-holdem', config={'allow_step_back':True})
        _, player_id = env.reset()
        env.step(0)
        _, back_player_id = env.step_back()
        self.assertEqual(player_id, back_player_id)
        self.assertEqual(env.step_back(), False)

        env = rlcard.make('no-limit-holdem')
        with self.assertRaises(Exception):
            env.step_back()

    def test_run(self):
        env = rlcard.make('no-limit-holdem')
        agents = [RandomAgent(env.num_actions) for _ in range(env.num_players)]
        env.set_agents(agents)
        trajectories, payoffs = env.run(is_training=False)
        self.assertEqual(len(trajectories), 2)
        total = 0
        for payoff in payoffs:
            total += payoff
        self.assertEqual(total, 0)

    def test_get_perfect_information(self):
        env = rlcard.make('no-limit-holdem')
        _, player_id = env.reset()
        self.assertEqual(player_id, env.get_perfect_information()['current_player'])

    def test_multiplayers(self):
        env = rlcard.make('no-limit-holdem', config={'game_num_players':5})
        num_players = env.game.get_num_players()
        self.assertEqual(num_players, 5)

    def test_config_chips(self):
        env = rlcard.make('no-limit-holdem', config={'game_num_players':5, 'chips_for_each':100})
        env.game.init_game()
        players = env.game.players
        chips = []
        for i in range(5):
            chips.append(players[i].remained_chips + players[i].in_chips)
        self.assertEqual(chips, [100, 100, 100, 100, 100])

if __name__ == '__main__':
    unittest.main()



from rlcard.utils.utils import print_card


class HumanAgent(object):
    ''' A human agent for Leduc Holdem. It can be used to play against trained models
    '''

    def __init__(self, num_actions):
        ''' Initilize the human agent

        Args:
            num_actions (int): the size of the ouput action space
        '''
        self.use_raw = True
        self.num_actions = num_actions

    @staticmethod
    def step(state):
        ''' Human agent will display the state and make decisions through interfaces

        Args:
            state (dict): A dictionary that represents the current state

        Returns:
            action (int): The action decided by human
        '''
        _print_state(state['raw_obs'], state['action_record'])
        action = int(input('>> You choose action (integer): '))
        while action < 0 or action >= len(state['legal_actions']):
            print('Action illegal...')
            action = int(input('>> Re-choose action (integer): '))
        return state['raw_legal_actions'][action]

    def eval_step(self, state):
        ''' Predict the action given the curent state for evaluation. The same to step here.

        Args:
            state (numpy.array): an numpy array that represents the current state

        Returns:
            action (int): the action predicted (randomly chosen) by the random agent
        '''
        return self.step(state), {}

def _print_state(state, action_record):
    ''' Print out the state

    Args:
        state (dict): A dictionary of the raw state
        action_record (list): A list of the historical actions
    '''
    _action_list = []
    for i in range(1, len(action_record)+1):
        if action_record[-i][0] == state['current_player']:
            break
        _action_list.insert(0, action_record[-i])
    for pair in _action_list:
        print('>> Player', pair[0], 'chooses', pair[1])

    print('\n=============== Community Card ===============')
    print_card(state['public_card'])
    print('===============   Your Hand    ===============')
    print_card(state['hand'])
    print('===============     Chips      ===============')
    print('Yours:   ', end='')
    for _ in range(state['my_chips']):
        print('+', end='')
    print('')
    for i in range(len(state['all_chips'])):
        if i != state['current_player']:
            print('Agent {}: '.format(i) , end='')
            for _ in range(state['all_chips'][i]):
                print('+', end='')
    print('\n=========== Actions You Can Choose ===========')
    print(', '.join([str(index) + ': ' + action for index, action in enumerate(state['legal_actions'])]))
    print('')



from rlcard.games.blackjack.dealer import BlackjackDealer as Dealer
from rlcard.games.blackjack.judger import BlackjackJudger as Judger
from rlcard.games.blackjack.player import BlackjackPlayer as Player
from rlcard.games.blackjack.game import BlackjackGame as Game




import unittest

from rlcard.agents.human_agents.leduc_holdem_human_agent import _print_state

class TestLeducHuman(unittest.TestCase):

    def test_print_state(self):
        raw_state = {'my_chips': 1, 'current_player': 0, 'all_chips': [1, 1], 'public_card': None, 'hand': 'SQ', 'legal_actions': ['raise', 'fold', 'check']}
        action_record = []
        _print_state(raw_state, action_record)




import unittest
import numpy as np
from rlcard.utils.utils import init_54_deck, init_standard_deck, rank2int, print_card, elegent_form, reorganize, tournament
import rlcard
from rlcard.agents.random_agent import RandomAgent

class TestUtils(unittest.TestCase):

    def test_init_standard_deck(self):
        self.assertEqual(len(init_standard_deck()), 52)

    def test_init_54_deck(self):
        self.assertEqual(len(init_54_deck()), 54)

    def test_rank2int(self):
        self.assertEqual(rank2int('A'), 14)
        self.assertEqual(rank2int(''), -1)
        self.assertEqual(rank2int('3'), 3)
        self.assertEqual(rank2int('T'), 10)
        self.assertEqual(rank2int('J'), 11)
        self.assertEqual(rank2int('Q'), 12)
        self.assertEqual(rank2int('1000'), None)
        self.assertEqual(rank2int('abc123'), None)
        self.assertEqual(rank2int('K'), 13)

    def test_print_cards(self):
        self.assertEqual(len(elegent_form('S9')), 2)
        self.assertEqual(len(elegent_form('ST')), 3)

        print_card(None)
        print_card('S9')
        print_card('ST')

    def test_reorganize(self):
        trajectories = reorganize([[[1,2],1,[4,5]]], [1])
        self.assertEqual(len(trajectories), 1)
        self.assertEqual(len(trajectories[0]), 1)
        self.assertEqual(len(trajectories[0][0]), 5)

    def test_tournament(self):
        env = rlcard.make('leduc-holdem')
        env.set_agents([RandomAgent(env.num_actions), RandomAgent(env.num_actions)])
        payoffs = tournament(env,1000)
        self.assertEqual(len(payoffs), 2)

if __name__ == '__main__':
    unittest.main()



'''
    File name: test_gin_rummy_game.py
    Author: William Hale
    Date created: 3/11/2020
'''

import unittest
import numpy as np

import rlcard.games.gin_rummy.judge as judge
import rlcard.games.gin_rummy.utils.utils as utils

from rlcard.games.gin_rummy.dealer import GinRummyDealer
from rlcard.games.gin_rummy.game import GinRummyGame as Game
from rlcard.games.gin_rummy.player import GinRummyPlayer
from rlcard.games.gin_rummy.utils.action_event import score_player_1_action_id
from rlcard.games.gin_rummy.utils.action_event import draw_card_action_id, pick_up_discard_action_id
from rlcard.games.gin_rummy.utils.action_event import declare_dead_hand_action_id
from rlcard.games.gin_rummy.utils.action_event import gin_action_id, discard_action_id, knock_action_id
from rlcard.games.gin_rummy.utils.melding import get_all_set_melds, get_all_run_melds, get_meld_clusters
from rlcard.games.gin_rummy.utils.settings import Setting, Settings
from rlcard.games.gin_rummy.utils.thinker import Thinker

discard_action_ids = list(range(discard_action_id, discard_action_id + 52))
knock_action_ids = list(range(knock_action_id, knock_action_id + 52))
put_action_ids = [gin_action_id] + discard_action_ids + knock_action_ids
get_action_ids = [draw_card_action_id, pick_up_discard_action_id, declare_dead_hand_action_id]


class TestGinRummyGame(unittest.TestCase):

    def test_get_num_players(self):
        game = Game()
        num_players = game.get_num_players()
        self.assertEqual(num_players, 2)

    def test_get_num_actions(self):
        game = Game()
        num_actions = game.get_num_actions()
        self.assertEqual(num_actions, 110)

    def test_init_game(self):
        game = Game()
        state, current_player = game.init_game()
        opponent_player = (current_player + 1) % 2
        self.assertEqual(len(game.round.move_sheet), 1)
        self.assertIn(current_player, [0, 1])
        self.assertIn(game.round.dealer_id, [0, 1])
        self.assertEqual(len(game.actions), 0)
        self.assertEqual(opponent_player, game.round.dealer_id)  # opponent_player is dealer
        self.assertEqual(len(game.round.players[opponent_player].hand), 10)  # dealer has 10 cards
        self.assertEqual(len(game.round.players[current_player].hand), 11)  # current_player has 11 cards
        self.assertEqual(len(game.round.dealer.shuffled_deck), 52)
        self.assertEqual(len(game.round.dealer.stock_pile), 31)
        self.assertEqual(state['player_id'], current_player)
        self.assertEqual(len(state['hand']), 11)

    def test_step(self):
        game = Game()
        _, current_player = game.init_game()
        opponent_player = (current_player + 1) % 2
        action = np.random.choice(game.judge.get_legal_actions())
        self.assertIn(action.action_id, put_action_ids)  # should be a put action
        _, next_player = game.step(action)
        if not game.is_over():
            self.assertEqual(next_player, opponent_player)
        if not game.is_over():
            action = np.random.choice(game.judge.get_legal_actions())
            self.assertIn(action.action_id, get_action_ids)  # should be a get action
            _, next_player = game.step(action)
            self.assertEqual(next_player, opponent_player)  # keep turn to put card

    def test_proceed_game(self):
        game = Game()
        game.init_game()
        while not game.is_over():
            legal_actions = game.judge.get_legal_actions()
            action = np.random.choice(legal_actions)
            _, _ = game.step(action)
        self.assertEqual(game.actions[-1].action_id, score_player_1_action_id)

    def test_get_state(self):
        game = Game()
        state, _ = game.init_game()
        self.assertEqual(len(state), 6)

    def test_gin_rummy_dealer(self):
        dealer = GinRummyDealer(np.random.RandomState())
        current_deck = utils.get_deck()
        deck_card_ids = [utils.get_card_id(card) for card in current_deck]
        self.assertEqual(deck_card_ids, list(range(52)))
        # Deal 10 cards.
        player = GinRummyPlayer(player_id=0, np_random=np.random.RandomState())
        dealer.deal_cards(player=player, num=10)
        self.assertEqual(len(dealer.shuffled_deck), 52)
        self.assertEqual(len(dealer.stock_pile), 42)
        self.assertEqual(len(current_deck), 52)
        self.assertEqual(len(utils.get_deck()), 52)
        # Pop top_card from current_deck.
        top_card = current_deck.pop(-1)
        self.assertEqual(str(top_card), "KC")
        self.assertEqual(len(current_deck), 51)
        self.assertEqual(len(utils.get_deck()), 52)

    def test_knocking(self):
        hand_text = ['JS', 'JH', 'JD', '8C', '7S', '7H', '7D', '4S', '3D', '2S', 'AC']
        hand = [utils.card_from_text(x) for x in hand_text]
        knock_cards, _ = judge.get_going_out_cards(hand=hand, going_out_deadwood_count=10)
        self.assertEqual(set(knock_cards), set([utils.card_from_text(x) for x in ['8C']]))

    def test_melding(self):
        hand_text = ['9H', 'AC', 'TH', '3C', '3D', '7C', 'QH', '3H', '8C', '8D',
                     '4D', '7H', '8S', '5H', '4H', 'AS', 'TD', '3S', '2S', 'AH']
        hand = [utils.card_from_text(x) for x in hand_text]

        # check
        all_set_melds_text = [['3C', '3D', '3H', '3S'], ['8C', '8D', '8S'], ['AC', 'AS', 'AH'], ['3D', '3H', '3S'],
                              ['3C', '3H', '3S'], ['3C', '3D', '3S'], ['3C', '3D', '3H']]
        all_set_melds_text = set([frozenset(x) for x in all_set_melds_text])
        all_set_melds = get_all_set_melds(hand)
        self.assertEqual(all_set_melds_text,
                         set([frozenset([str(card) for card in meld_pile]) for meld_pile in all_set_melds]))

        # check
        all_run_melds_text = [['AS', '2S', '3S'], ['3H', '4H', '5H']]
        all_run_melds = get_all_run_melds(hand)
        self.assertEqual(all_run_melds_text, [[str(card) for card in meld_pile] for meld_pile in all_run_melds])

        # check
        meld_clusters = get_meld_clusters(hand=hand)
        self.assertEqual(len(meld_clusters), 36)

    def test_melding_2(self):  # 2020-Apr-19
        hand_text = ['8D', '5S', '5H', '5D', '4H', '4C', '3H', '2H', '2D', '2C', 'AD']
        hand = [utils.card_from_text(x) for x in hand_text]
        going_out_deadwood_count = 10
        knock_cards, gin_cards = judge.get_going_out_cards(hand=hand, going_out_deadwood_count=going_out_deadwood_count)

        player = GinRummyPlayer(player_id=0, np_random=np.random.RandomState())
        player.hand = hand
        player.did_populate_hand()
        meld_clusters = player.get_meld_clusters()
        alpha, beta = judge._get_going_out_cards(meld_clusters=meld_clusters,
                                                 hand=hand,
                                                 going_out_deadwood_count=going_out_deadwood_count)
        self.assertEqual(set(alpha), set(knock_cards))
        self.assertEqual(beta, [])
        self.assertEqual(knock_cards, [utils.card_from_text('8D')])
        self.assertEqual(gin_cards, [])

    def test_melding_3(self):  # 2020-Apr-19
        hand_text = ['7H', '6H', '5H', '4S', '4H', '3H', '2S', 'AS', 'AH', 'AD', 'AC']
        hand = [utils.card_from_text(x) for x in hand_text]
        going_out_deadwood_count = 10
        knock_cards, gin_cards = judge.get_going_out_cards(hand=hand, going_out_deadwood_count=going_out_deadwood_count)

        player = GinRummyPlayer(player_id=0, np_random=np.random.RandomState())
        player.hand = hand
        player.did_populate_hand()
        meld_clusters = player.get_meld_clusters()
        alpha, beta = judge._get_going_out_cards(meld_clusters=meld_clusters,
                                                 hand=hand,
                                                 going_out_deadwood_count=going_out_deadwood_count)
        self.assertEqual(set(alpha), set(knock_cards))
        self.assertEqual(beta, [])

        correct_knock_cards = [utils.card_from_text(x) for x in ['7H', '4S', '4H', '3H', '2S', 'AS', 'AH', 'AD', 'AC']]
        self.assertEqual(set(knock_cards), set(correct_knock_cards))
        self.assertEqual(gin_cards, [])

    def test_corrected_settings(self):
        default_setting = Setting.default_setting()
        config = {Setting.max_drawn_card_count: 10,
                  Setting.stockpile_dead_card_count: 4.5,
                  Setting.is_allowed_pick_up_discard: 0}
        corrected_config = Settings.get_config_with_invalid_settings_set_to_default_value(config=config)
        self.assertEqual(corrected_config[Setting.max_drawn_card_count], 10)
        self.assertEqual(corrected_config[Setting.stockpile_dead_card_count],
                         default_setting[Setting.stockpile_dead_card_count])
        self.assertEqual(corrected_config[Setting.is_allowed_pick_up_discard],
                         default_setting[Setting.is_allowed_pick_up_discard])

    def test_change_settings(self):
        default_setting = Setting.default_setting()
        config = {Setting.max_drawn_card_count: 10,
                  Setting.stockpile_dead_card_count: 4.5,
                  Setting.is_allowed_pick_up_discard: 0}
        settings = Settings()
        settings.change_settings(config=config)
        self.assertEqual(settings.max_drawn_card_count, 10)
        self.assertEqual(settings.stockpile_dead_card_count,
                         default_setting[Setting.stockpile_dead_card_count])
        self.assertEqual(settings.is_allowed_pick_up_discard,
                         default_setting[Setting.is_allowed_pick_up_discard])

    def test_decode_cards(self):
        deck = utils.get_deck()
        encoded_cards = utils.encode_cards(deck)
        decoded_cards = utils.decode_cards(encoded_cards)
        for i in range(52):
            card = deck[i]
            decoded_card = decoded_cards[i]
            self.assertEqual(card, decoded_card)

    def test_get_meld_piles_with_discard_card(self):
        hand_text = ['8D', '7D', '6S', '5H', '5C', '4C', '4S', '2S', 'AC', 'AH']
        hand = [utils.card_from_text(x) for x in hand_text]
        discard_card = utils.card_from_text('6D')
        thinker = Thinker(hand=hand)
        meld_piles_with_discard_card = thinker.get_meld_piles_with_discard_card(discard_card=discard_card)
        result_as_set = frozenset([frozenset(meld_pile) for meld_pile in meld_piles_with_discard_card])
        correct_result = [[utils.card_from_text(x) for x in ['8D', '7D', '6D']]]
        correct_result_as_set = frozenset([frozenset(meld_pile) for meld_pile in correct_result])
        self.assertEqual(result_as_set, correct_result_as_set)


if __name__ == '__main__':
    unittest.main()



import subprocess
import sys
from distutils.version import LooseVersion

reqs = subprocess.check_output([sys.executable, '-m', 'pip', 'freeze'])
installed_packages = [r.decode().split('==')[0] for r in reqs.split()]

if 'torch' in installed_packages:
    from rlcard.agents.dqn_agent import DQNAgent as DQNAgent
    from rlcard.agents.nfsp_agent import NFSPAgent as NFSPAgent

from rlcard.agents.cfr_agent import CFRAgent
from rlcard.agents.human_agents.limit_holdem_human_agent import HumanAgent as LimitholdemHumanAgent
from rlcard.agents.human_agents.nolimit_holdem_human_agent import HumanAgent as NolimitholdemHumanAgent
from rlcard.agents.human_agents.leduc_holdem_human_agent import HumanAgent as LeducholdemHumanAgent
from rlcard.agents.human_agents.blackjack_human_agent import HumanAgent as BlackjackHumanAgent
from rlcard.agents.human_agents.uno_human_agent import HumanAgent as UnoHumanAgent
from rlcard.agents.random_agent import RandomAgent



'''
    File name: test_bridge_game.py
    Author: William Hale
    Date created: 11/25/2021
'''

import unittest
import numpy as np

from rlcard.games.bridge.game import BridgeGame as Game
from rlcard.games.bridge.dealer import BridgeDealer
from rlcard.games.bridge.player import BridgePlayer
from rlcard.games.bridge.utils.action_event import PassAction
from rlcard.games.bridge.utils.bridge_card import BridgeCard
from rlcard.games.bridge.utils.move import DealHandMove


class TestBridgeGame(unittest.TestCase):

    def test_get_num_players(self):
        game = Game()
        num_players = game.get_num_players()
        self.assertEqual(num_players, 4)

    def test_get_num_actions(self):
        game = Game()
        num_actions = game.get_num_actions()
        self.assertEqual(num_actions, 91)

    def test_bridge_dealer(self):
        dealer = BridgeDealer(np.random.RandomState())
        current_deck = BridgeCard.get_deck()
        deck_card_ids = [card.card_id for card in current_deck]
        self.assertEqual(deck_card_ids, list(range(52)))
        # Deal 13 cards.
        player = BridgePlayer(player_id=0, np_random=np.random.RandomState())
        dealer.deal_cards(player=player, num=13)
        self.assertEqual(len(dealer.shuffled_deck), 52)
        self.assertEqual(len(dealer.stock_pile), 39)
        self.assertEqual(len(current_deck), 52)
        self.assertEqual(len(BridgeCard.get_deck()), 52)
        # Pop top_card from current_deck.
        top_card = current_deck.pop(-1)
        self.assertEqual(str(top_card), "AS")
        self.assertEqual(len(current_deck), 51)
        self.assertEqual(len(BridgeCard.get_deck()), 52)

    def test_init_game(self):
        player_ids = list(range(4))
        game = Game()
        state, current_player = game.init_game()
        self.assertEqual(len(game.round.move_sheet), 1)
        self.assertIn(current_player, player_ids)
        self.assertEqual(len(game.actions), 0)
        self.assertEqual(len(game.round.players[current_player].hand), 13)  # current_player has 13 cards
        self.assertEqual(len(game.round.dealer.shuffled_deck), 52)
        self.assertEqual(len(game.round.dealer.stock_pile), 0)
        self.assertEqual(state['player_id'], current_player)
        self.assertEqual(len(state['hand']), 13)

    def test_step(self):
        game = Game()
        _, current_player_id = game.init_game()
        legal_actions = game.judger.get_legal_actions()
        action = np.random.choice(legal_actions)
        print(f'test_step current_player_id={current_player_id} action={action} legal_actions={[str(action) for action in legal_actions]}')
        _, next_player_id = game.step(action)
        next_legal_actions = game.judger.get_legal_actions()
        print(f'test_step next_player_id={next_player_id} next_legal_actions={[str(action) for action in next_legal_actions]}')

    def test_first_legal_bids(self):
        game = Game()
        _, current_player_id = game.init_game()
        for _ in range(3):
            legal_actions = game.judger.get_legal_actions()
            self.assertEqual(len(legal_actions), 36)
            action = PassAction()
            self.assertTrue(action in legal_actions)
            _, next_player_id = game.step(action)
        legal_actions = game.judger.get_legal_actions()
        self.assertEqual(len(legal_actions), 36)
        self.assertTrue(PassAction() in legal_actions)

    def test_pass_out_hand(self):
        game = Game()
        _, current_player_id = game.init_game()
        for _ in range(4):
            legal_actions = game.judger.get_legal_actions()
            action = PassAction()
            self.assertTrue(action in legal_actions)
            _, next_player_id = game.step(action)
        self.assertTrue(not game.round.contract_bid_move)
        self.assertTrue(game.round.is_bidding_over())
        self.assertTrue(game.is_over())
        self.assertTrue(not game.round.get_declarer())
        self.assertTrue(not game.round.get_dummy())
        self.assertTrue(not game.round.get_left_defender())
        self.assertTrue(not game.round.get_right_defender())

    def test_play_game(self):
        game = Game()
        next_state, next_player_id = game.init_game()
        deal_hand_move = game.round.move_sheet[0]
        self.assertTrue(isinstance(deal_hand_move, DealHandMove))
        print(f'test_play_game {deal_hand_move}')
        while not game.is_over():
            current_player_id = game.round.current_player_id
            self.assertEqual(current_player_id, next_player_id)
            legal_actions = game.judger.get_legal_actions()
            action = np.random.choice(legal_actions)
            print(f'test_play_game {current_player_id} {action} from {[str(x) for x in legal_actions]}')
            next_state, next_player_id = game.step(action)
        for player_id in range(4):
            player = game.round.players[player_id]
            hand = player.hand
            self.assertTrue(not hand)

    def test_print_scene(self):
        game = Game()
        next_state, next_player_id = game.init_game()
        deal_hand_move = game.round.move_sheet[0]
        self.assertTrue(isinstance(deal_hand_move, DealHandMove))
        while not game.is_over():
            current_player_id = game.round.current_player_id
            self.assertEqual(current_player_id, next_player_id)
            legal_actions = game.judger.get_legal_actions()
            action = np.random.choice(legal_actions)
            game.round.print_scene()
            next_state, next_player_id = game.step(action)
        game.round.print_scene()
        for player_id in range(4):
            player = game.round.players[player_id]
            hand = player.hand
            self.assertTrue(not hand)


if __name__ == '__main__':
    unittest.main()



from rlcard.agents.nfsp_agent import NFSPAgent
from rlcard.agents.dqn_agent import DQNAgent
from rlcard.agents.random_agent import RandomAgent
from rlcard.utils.pettingzoo_utils import wrap_state


class NFSPAgentPettingZoo(NFSPAgent):
    def step(self, state):
        return super().step(wrap_state(state))

    def eval_step(self, state):
        return super().eval_step(wrap_state(state))

    def feed(self, ts):
        state, action, reward, next_state, done = tuple(ts)
        state = wrap_state(state)
        next_state = wrap_state(next_state)
        ts = (state, action, reward, next_state, done)
        return super().feed(ts)


class DQNAgentPettingZoo(DQNAgent):
    def step(self, state):
        return super().step(wrap_state(state))

    def eval_step(self, state):
        return super().eval_step(wrap_state(state))

    def feed(self, ts):
        state, action, reward, next_state, done = tuple(ts)
        state = wrap_state(state)
        next_state = wrap_state(next_state)
        ts = (state, action, reward, next_state, done)
        return super().feed(ts)


class RandomAgentPettingZoo(RandomAgent):
    def step(self, state):
        return super().step(wrap_state(state))

    def eval_step(self, state):
        return super().eval_step(wrap_state(state))



import unittest

from rlcard.models.model import Model
from rlcard.models.pretrained_models import LeducHoldemCFRModel
from rlcard.models.leducholdem_rule_models import LeducHoldemRuleModelV1, LeducHoldemRuleModelV2

from rlcard.models.limitholdem_rule_models import LimitholdemRuleModelV1
from rlcard.models.doudizhu_rule_models import DouDizhuRuleModelV1

from rlcard.models.gin_rummy_rule_models import GinRummyNoviceRuleModel


class TestModel(unittest.TestCase):

    def test_model(self):
        model = Model()
        self.assertIsInstance(model, Model)

    def test_leduc_holdem_cfr_model(self):
        model = LeducHoldemCFRModel()
        self.assertIsInstance(model, LeducHoldemCFRModel)
        self.assertIsInstance(model.agents, list)

    def test_leduc_holdem_rule_model_v1(self):
        model = LeducHoldemRuleModelV1()
        self.assertIsInstance(model, LeducHoldemRuleModelV1)
        agent = model.agents[0]
        action = agent.step({'raw_legal_actions':['raise']})
        self.assertEqual(action, 'raise')
        action = agent.step({'raw_legal_actions':['call']})
        self.assertEqual(action, 'call')
        action = agent.step({'raw_legal_actions':['check']})
        self.assertEqual(action, 'check')
        action = agent.step({'raw_legal_actions':[]})
        self.assertEqual(action, 'fold')

    def test_leduc_holdem_rule_model_v2(self):
        model = LeducHoldemRuleModelV2()
        self.assertIsInstance(model, LeducHoldemRuleModelV2)
        agent = model.agents[0]
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['K'], 'public_card':[]}})
        self.assertEqual(action, 'raise')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['Q'], 'public_card':[]}})
        self.assertEqual(action, 'check')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['J'], 'public_card':[]}})
        self.assertEqual(action, 'fold')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['K', 'J'], 'public_card':['K','J']}})
        self.assertEqual(action, 'raise')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['K', 'Q'], 'public_card':['K','J']}})
        self.assertEqual(action, 'fold')
        action = agent.step({'raw_legal_actions':['fold', 'check', 'call'], 'raw_obs':{'hand':['K'], 'public_card':[]}})
        self.assertEqual(action, 'call')
        action = agent.step({'raw_legal_actions':['fold', 'call'], 'raw_obs':{'hand':['Q'], 'public_card':[]}})
        self.assertEqual(action, 'fold')

    def test_limit_holdem_rule_model_v1(self):
        model = LimitholdemRuleModelV1()
        self.assertIsInstance(model, LimitholdemRuleModelV1)
        agent = model.agents[0]
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['S2', 'H4'], 'public_cards':[]}})
        self.assertEqual(action, 'fold')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['SA', 'HA'], 'public_cards':[]}})
        self.assertEqual(action, 'raise')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['SA', 'HT'], 'public_cards':[]}})
        self.assertEqual(action, 'raise')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['S2', 'SA'], 'public_cards':[]}})
        self.assertEqual(action, 'raise')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['HQ', 'SJ'], 'public_cards':[]}})
        self.assertEqual(action, 'raise')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['HQ', 'S2'], 'public_cards':[]}})
        self.assertEqual(action, 'fold')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['SA', 'HA'], 'public_cards':['CA', 'C2', 'B4']}})
        self.assertEqual(action, 'raise')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['SA', 'HQ'], 'public_cards':['CJ', 'C2', 'B4']}})
        self.assertEqual(action, 'raise')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['HA', 'H9'], 'public_cards':['HJ', 'C2', 'B4']}})
        self.assertEqual(action, 'raise')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['SK', 'HQ'], 'public_cards':['H6', 'C2', 'B4']}})
        self.assertEqual(action, 'call')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['SK', 'HQ'], 'public_cards':['H2', 'C2', 'B4']}})
        self.assertEqual(action, 'check')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['SA', 'HA'], 'public_cards':['CA', 'C2', 'B4', 'B6']}})
        self.assertEqual(action, 'raise')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['SA', 'HQ'], 'public_cards':['CJ', 'C2', 'B4', 'B6']}})
        self.assertEqual(action, 'raise')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['H9', 'HA'], 'public_cards':['HJ', 'C2', 'B4', 'B6']}})
        self.assertEqual(action, 'raise')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['H9', 'HA'], 'public_cards':['HJ', 'C2', 'B4', 'B6']}})
        self.assertEqual(action, 'raise')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['SK', 'HQ'], 'public_cards':['H6', 'C2', 'B4', 'B5']}})
        self.assertEqual(action, 'call')
        action = agent.step({'raw_legal_actions':['raise', 'fold', 'check', 'call'], 'raw_obs':{'hand':['SK', 'HQ'], 'public_cards':['H2', 'C2', 'B4', 'B5']}})
        self.assertEqual(action, 'fold')

    def test_doudizhu_rule_model_v1(self):
        model = DouDizhuRuleModelV1()
        self.assertIsInstance(model, DouDizhuRuleModelV1)
        agent = model.agents[0]
        action = agent.step({'raw_obs': {'current_hand': '345567999TTTTQKAA2BR', 'trace':[]}})
        self.assertEqual(action, '34567')
        action = agent.step({'raw_obs': {'current_hand': '5999TTTTQKAA2BR', 'trace':[(0, '34567'), (1, 'pass'), (2, 'pass')]}})
        self.assertEqual(action, '5')
        action = agent.step({'raw_obs': {'actions': ['pass', '44455', '44466', 'TTTT', 'BR'], 'trace':[(0, '33344')]}})
        self.assertEqual(action, '44455')
        action = agent.step({'raw_obs': {'actions': ['pass', '44455', '44466', 'TTTT', 'BR'], 'trace':[(0, '33344'), [1, 'pass']]}})
        self.assertEqual(action, '44455')
        action = agent.step({'raw_obs': {'actions': ['pass', 'TTTT', 'BR'], 'self': 2, 'landlord': 0, 'trace':[(0, '33344'), (1, '55566')]}})
        self.assertEqual(action, 'pass')

    def test_gin_rummy_novice_model(self):
        model = GinRummyNoviceRuleModel()
        self.assertIsInstance(model, GinRummyNoviceRuleModel)
        self.assertIsInstance(model.agents, list)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 RLCard Team of Texas A&M University
# Copyright 2021 DouZero Team of Kwai
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import traceback

import numpy as np
import torch

shandle = logging.StreamHandler()
shandle.setFormatter(
    logging.Formatter(
        '[%(levelname)s:%(process)d %(module)s:%(lineno)d %(asctime)s] '
        '%(message)s'))
log = logging.getLogger('doudzero')
log.propagate = False
log.addHandler(shandle)
log.setLevel(logging.INFO)

def get_batch(
    free_queue,
    full_queue,
    buffers,
    batch_size,
    lock
):
    with lock:
        indices = [full_queue.get() for _ in range(batch_size)]
    batch = {
        key: torch.stack([buffers[key][m] for m in indices], dim=1)
        for key in buffers
    }
    for m in indices:
        free_queue.put(m)
    return batch

def create_buffers(
    T,
    num_buffers,
    state_shape,
    action_shape,
    device_iterator,
):
    buffers = {}
    for device in device_iterator:
        buffers[device] = []
        for player_id in range(len(state_shape)):
            specs = dict(
                done=dict(size=(T,), dtype=torch.bool),
                episode_return=dict(size=(T,), dtype=torch.float32),
                target=dict(size=(T,), dtype=torch.float32),
                state=dict(size=(T,)+tuple(state_shape[player_id]), dtype=torch.int8),
                action=dict(size=(T,)+tuple(action_shape[player_id]), dtype=torch.int8),
            )
            _buffers = {key: [] for key in specs}
            for _ in range(num_buffers):
                for key in _buffers:
                    if device == "cpu":
                        _buffer = torch.empty(**specs[key]).to('cpu').share_memory_()
                    else:
                        _buffer = torch.empty(**specs[key]).to('cuda:'+str(device)).share_memory_()
                    _buffers[key].append(_buffer)
            buffers[device].append(_buffers)
    return buffers

def create_optimizers(
    num_players,
    learning_rate,
    momentum,
    epsilon,
    alpha,
    learner_model
):
    optimizers = []
    for player_id in range(num_players):
        optimizer = torch.optim.RMSprop(
            learner_model.parameters(player_id),
            lr=learning_rate,
            momentum=momentum,
            eps=epsilon,
            alpha=alpha)
        optimizers.append(optimizer)
    return optimizers

def act(
    i,
    device,
    T,
    free_queue,
    full_queue,
    model,
    buffers,
    env
):
    try:
        log.info('Device %s Actor %i started.', str(device), i)

        # Configure environment
        env.seed(i)
        env.set_agents(model.get_agents())

        done_buf = [[] for _ in range(env.num_players)]
        episode_return_buf = [[] for _ in range(env.num_players)]
        target_buf = [[] for _ in range(env.num_players)]
        state_buf = [[] for _ in range(env.num_players)]
        action_buf = [[] for _ in range(env.num_players)]
        size = [0 for _ in range(env.num_players)]

        while True:
            trajectories, payoffs = env.run(is_training=True)
            for p in range(env.num_players):
                size[p] += len(trajectories[p][:-1]) // 2
                diff = size[p] - len(target_buf[p])
                if diff > 0:
                    done_buf[p].extend([False for _ in range(diff-1)])
                    done_buf[p].append(True)
                    episode_return_buf[p].extend([0.0 for _ in range(diff-1)])
                    episode_return_buf[p].append(float(payoffs[p]))
                    target_buf[p].extend([float(payoffs[p]) for _ in range(diff)])
                    # State and action
                    for i in range(0, len(trajectories[p])-2, 2):
                        state = trajectories[p][i]['obs']
                        action = env.get_action_feature(trajectories[p][i+1])
                        state_buf[p].append(torch.from_numpy(state))
                        action_buf[p].append(torch.from_numpy(action))
                
                while size[p] > T:
                    index = free_queue[p].get()
                    if index is None:
                        break
                    for t in range(T):
                        buffers[p]['done'][index][t, ...] = done_buf[p][t]
                        buffers[p]['episode_return'][index][t, ...] = episode_return_buf[p][t]
                        buffers[p]['target'][index][t, ...] = target_buf[p][t]
                        buffers[p]['state'][index][t, ...] = state_buf[p][t]
                        buffers[p]['action'][index][t, ...] = action_buf[p][t]
                    full_queue[p].put(index)
                    done_buf[p] = done_buf[p][T:]
                    episode_return_buf[p] = episode_return_buf[p][T:]
                    target_buf[p] = target_buf[p][T:]
                    state_buf[p] = state_buf[p][T:]
                    action_buf[p] = action_buf[p][T:]
                    size[p] -= T

    except KeyboardInterrupt:
        pass
    except Exception as e:
        log.error('Exception in worker process %i', i)
        traceback.print_exc()
        print()
        raise e



'''
    Project: Gui Gin Rummy
    File name: game_canvas_debug.py
    Author: William Hale
    Date created: 3/14/2020
'''

# from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .game_canvas import GameCanvas

from . import configurations

from rlcard.games.gin_rummy.player import GinRummyPlayer

import rlcard.games.gin_rummy.utils.utils as gin_rummy_utils


class GameCanvasDebug(object):

    def __init__(self, game_canvas: 'GameCanvas'):
        self.game_canvas = game_canvas

    def get_card_name(self, card_item_id: int) -> str:
        card_id = self.game_canvas.card_item_ids.index(card_item_id)
        card = gin_rummy_utils.card_from_card_id(card_id=card_id)
        return str(card)

    def description(self):
        game_canvas = self.game_canvas
        card_name = self.get_card_name
        dealer_id = game_canvas.dealer_id
        current_player_id = game_canvas.current_player_id
        stock_pile_item_ids = game_canvas.find_withtag(configurations.STOCK_PILE_TAG)
        discard_pile_items = game_canvas.find_withtag(configurations.DISCARD_PILE_TAG)
        north_held_pile_item_ids = game_canvas.getter.get_held_pile_item_ids(player_id=0)
        south_held_pile_item_ids = game_canvas.getter.get_held_pile_item_ids(player_id=1)
        lines = []
        lines.append("dealer: {}".format(GinRummyPlayer.short_name_of(player_id=dealer_id)))
        lines.append("current_player: {}".format(GinRummyPlayer.short_name_of(player_id=current_player_id)))
        lines.append("north hand: {}".format([card_name(card_item_id) for card_item_id in north_held_pile_item_ids]))
        lines.append("stockpile: {}".format([card_name(card_item_id) for card_item_id in stock_pile_item_ids]))
        lines.append("discard pile: {}".format([card_name(card_item_id) for card_item_id in discard_pile_items]))
        lines.append("south hand: {}".format([card_name(card_item_id) for card_item_id in south_held_pile_item_ids]))
        return "\n".join(lines)



from rlcard.games.doudizhu.dealer import DoudizhuDealer as Dealer
from rlcard.games.doudizhu.judger import DoudizhuJudger as Judger
from rlcard.games.doudizhu.player import DoudizhuPlayer as Player
from rlcard.games.doudizhu.round import DoudizhuRound as Round
from rlcard.games.doudizhu.game import DoudizhuGame as Game




from collections import defaultdict
import numpy as np


def wrap_state(state):
    # check if obs is already wrapped
    if "obs" in state and "legal_actions" in state and "raw_legal_actions" in state:
        return state

    wrapped_state = {}
    wrapped_state["obs"] = state["observation"]
    legal_actions = np.flatnonzero(state["action_mask"])
    # the values of legal_actions isn't available so setting them to None
    wrapped_state["legal_actions"] = {l: None for l in legal_actions}
    # raw_legal_actions isn't available so setting it to legal actions
    wrapped_state["raw_legal_actions"] = list(wrapped_state["legal_actions"].keys())
    return wrapped_state


def run_game_pettingzoo(env, agents, is_training=False):
    env.reset()
    trajectories = defaultdict(list)
    for agent_name in env.agent_iter():
        obs, reward, done, _, _ = env.last()
        trajectories[agent_name].append((obs, reward, done))

        if done:
            action = None
        else:
            if is_training:
                action = agents[agent_name].step(obs)
            else:
                action, _ = agents[agent_name].eval_step(obs)
        trajectories[agent_name].append(action)

        env.step(action)
    return trajectories


def reorganize_pettingzoo(trajectories):
    ''' Reorganize the trajectory to make it RL friendly

    Args:
        trajectory (list): A list of trajectories

    Returns:
        (list): A new trajectories that can be fed into RL algorithms.

    '''
    new_trajectories = defaultdict(list)
    for agent_name, trajectory in trajectories.items():
        for i in range(0, len(trajectory)-2, 2):
            transition = [
                trajectory[i][0], # obs,
                trajectory[i+1], # action
                trajectory[i+2][1], # reward
                trajectory[i+2][0], # next_obs
                trajectory[i+2][2], # done
            ]
            new_trajectories[agent_name].append(transition)
    return new_trajectories


def tournament_pettingzoo(env, agents, num_episodes):
    total_rewards = defaultdict(float)
    for _ in range(num_episodes):
        trajectories = run_game_pettingzoo(env, agents)
        trajectories = reorganize_pettingzoo(trajectories)
        for agent_name, trajectory in trajectories.items():
            reward = sum([t[2] for t in trajectory])
            total_rewards[agent_name] += reward
    return {k: v / num_episodes for (k, v) in total_rewards.items()}



''' Leduc Hold 'em rule model
'''
import rlcard
from rlcard.models.model import Model

class LeducHoldemRuleAgentV1(object):
    ''' Leduc Hold 'em Rule agent version 1
    '''
    def __init__(self):
        self.use_raw = True

    @staticmethod
    def step(state):
        ''' Predict the action when given raw state. A simple rule-based AI.
        Args:
            state (dict): Raw state from the game

        Returns:
            action (str): Predicted action
        '''
        legal_actions = state['raw_legal_actions']
        # Aggressively play 'raise' and 'call'
        if 'raise' in legal_actions:
            return 'raise'
        if 'call' in legal_actions:
            return 'call'
        if 'check' in legal_actions:
            return 'check'
        else:
            return 'fold'

    def eval_step(self, state):
        ''' Step for evaluation. The same to step
        '''
        return self.step(state), []

class LeducHoldemRuleAgentV2(object):
    ''' Leduc Hold 'em Rule agent version 2
    '''
    def __init__(self):
        self.use_raw = True

    @staticmethod
    def step(state):
        ''' Predict the action when given raw state. A simple rule-based AI.
        Args:
            state (dict): Raw state from the game

        Returns:
            action (str): Predicted action
        '''
        legal_actions = state['raw_legal_actions']
        state = state['raw_obs']
        hand = state['hand']
        public_card = state['public_card']
        action = 'fold'
        # When having only 2 hand cards at the game start, choose fold to drop terrible cards:
        # Acceptable hand cards:
        # Pairs
        # AK, AQ, AJ, AT
        # A9s, A8s, ... A2s(s means flush)
        # KQ, KJ, QJ, JT
        # Fold all hand types except those mentioned above to save money
        if public_card:
            if public_card[1] == hand[1]:
                action = 'raise'
            else:
                action = 'fold'
        else:
            if hand[0] == 'K':
                action = 'raise'
            elif hand[0] == 'Q':
                action = 'check'
            else:
                action = 'fold'

        #return action
        if action in legal_actions:
            return action
        else:
            if action == 'raise':
                return 'call'
            if action == 'check':
                return 'fold'
            if action == 'call':
                return 'raise'
            else:
                return action

    def eval_step(self, state):
        ''' Step for evaluation. The same to step
        '''
        return self.step(state), []

class LeducHoldemRuleModelV1(Model):
    ''' Leduc holdem Rule Model version 1
    '''

    def __init__(self):
        ''' Load pretrained model
        '''
        env = rlcard.make('leduc-holdem')
        rule_agent = LeducHoldemRuleAgentV1()
        self.rule_agents = [rule_agent for _ in range(env.num_players)]

    @property
    def agents(self):
        ''' Get a list of agents for each position in a the game

        Returns:
            agents (list): A list of agents

        Note: Each agent should be just like RL agent with step and eval_step
              functioning well.
        '''
        return self.rule_agents

class LeducHoldemRuleModelV2(Model):
    ''' Leduc holdem Rule Model version 2
    '''

    def __init__(self):
        ''' Load pretrained model
        '''
        env = rlcard.make('leduc-holdem')
        rule_agent = LeducHoldemRuleAgentV2()
        self.rule_agents = [rule_agent for _ in range(env.num_players)]

    @property
    def agents(self):
        ''' Get a list of agents for each position in a the game

        Returns:
            agents (list): A list of agents

        Note: Each agent should be just like RL agent with step and eval_step
              functioning well.
        '''
        return self.rule_agents



'''
    File name: bridge/player.py
    Author: William Hale
    Date created: 11/25/2021
'''

from typing import List

from .utils.bridge_card import BridgeCard


class BridgePlayer:

    def __init__(self, player_id: int, np_random):
        ''' Initialize a BridgePlayer player class

        Args:
            player_id (int): id for the player
        '''
        if player_id < 0 or player_id > 3:
            raise Exception(f'BridgePlayer has invalid player_id: {player_id}')
        self.np_random = np_random
        self.player_id: int = player_id
        self.hand: List[BridgeCard] = []

    def remove_card_from_hand(self, card: BridgeCard):
        self.hand.remove(card)

    def __str__(self):
        return ['N', 'E', 'S', 'W'][self.player_id]



'''
    File name: bridge/game.py
    Author: William Hale
    Date created: 11/25/2021
'''

from typing import List

import numpy as np

from .judger import BridgeJudger
from .round import BridgeRound
from .utils.action_event import ActionEvent, CallActionEvent, PlayCardAction


class BridgeGame:
    ''' Game class. This class will interact with outer environment.
    '''

    def __init__(self, allow_step_back=False):
        '''Initialize the class BridgeGame
        '''
        self.allow_step_back: bool = allow_step_back
        self.np_random = np.random.RandomState()
        self.judger: BridgeJudger = BridgeJudger(game=self)
        self.actions: [ActionEvent] = []  # must reset in init_game
        self.round: BridgeRound or None = None  # must reset in init_game
        self.num_players: int = 4

    def init_game(self):
        ''' Initialize all characters in the game and start round 1
        '''
        board_id = self.np_random.choice([1, 2, 3, 4])
        self.actions: List[ActionEvent] = []
        self.round = BridgeRound(num_players=self.num_players, board_id=board_id, np_random=self.np_random)
        for player_id in range(4):
            player = self.round.players[player_id]
            self.round.dealer.deal_cards(player=player, num=13)
        current_player_id = self.round.current_player_id
        state = self.get_state(player_id=current_player_id)
        return state, current_player_id

    def step(self, action: ActionEvent):
        ''' Perform game action and return next player number, and the state for next player
        '''
        if isinstance(action, CallActionEvent):
            self.round.make_call(action=action)
        elif isinstance(action, PlayCardAction):
            self.round.play_card(action=action)
        else:
            raise Exception(f'Unknown step action={action}')
        self.actions.append(action)
        next_player_id = self.round.current_player_id
        next_state = self.get_state(player_id=next_player_id)
        return next_state, next_player_id

    def get_num_players(self) -> int:
        ''' Return the number of players in the game
        '''
        return self.num_players

    @staticmethod
    def get_num_actions() -> int:
        ''' Return the number of possible actions in the game
        '''
        return ActionEvent.get_num_actions()

    def get_player_id(self):
        ''' Return the current player that will take actions soon
        '''
        return self.round.current_player_id

    def is_over(self) -> bool:
        ''' Return whether the current game is over
        '''
        return self.round.is_over()

    def get_state(self, player_id: int):  # wch: not really used
        ''' Get player's state

        Return:
            state (dict): The information of the state
        '''
        state = {}
        if not self.is_over():
            state['player_id'] = player_id
            state['current_player_id'] = self.round.current_player_id
            state['hand'] = self.round.players[player_id].hand
        else:
            state['player_id'] = player_id
            state['current_player_id'] = self.round.current_player_id
            state['hand'] = self.round.players[player_id].hand
        return state



'''
    File name: bridge/utils/move.py
    Author: William Hale
    Date created: 11/25/2021
'''

#
#   These classes are used to keep a move_sheet history of the moves in a round.
#

from .action_event import ActionEvent, BidAction, PassAction, DblAction, RdblAction, PlayCardAction
from .bridge_card import BridgeCard

from ..player import BridgePlayer


class BridgeMove(object):  # Interface
    pass


class PlayerMove(BridgeMove):  # Interface

    def __init__(self, player: BridgePlayer, action: ActionEvent):
        super().__init__()
        self.player = player
        self.action = action


class CallMove(PlayerMove):  # Interface

    def __init__(self, player: BridgePlayer, action: ActionEvent):
        super().__init__(player=player, action=action)


class DealHandMove(BridgeMove):

    def __init__(self, dealer: BridgePlayer, shuffled_deck: [BridgeCard]):
        super().__init__()
        self.dealer = dealer
        self.shuffled_deck = shuffled_deck

    def __str__(self):
        shuffled_deck_text = " ".join([str(card) for card in self.shuffled_deck])
        return f'{self.dealer} deal shuffled_deck=[{shuffled_deck_text}]'


class MakePassMove(CallMove):

    def __init__(self, player: BridgePlayer):
        super().__init__(player=player, action=PassAction())

    def __str__(self):
        return f'{self.player} {self.action}'


class MakeDblMove(CallMove):

    def __init__(self, player: BridgePlayer):
        super().__init__(player=player, action=DblAction())

    def __str__(self):
        return f'{self.player} {self.action}'


class MakeRdblMove(CallMove):

    def __init__(self, player: BridgePlayer):
        super().__init__(player=player, action=RdblAction())

    def __str__(self):
        return f'{self.player} {self.action}'


class MakeBidMove(CallMove):

    def __init__(self, player: BridgePlayer, bid_action: BidAction):
        super().__init__(player=player, action=bid_action)
        self.action = bid_action  # Note: keep type as BidAction rather than ActionEvent

    def __str__(self):
        return f'{self.player} bids {self.action}'


class PlayCardMove(PlayerMove):

    def __init__(self, player: BridgePlayer, action: PlayCardAction):
        super().__init__(player=player, action=action)
        self.action = action  # Note: keep type as PlayCardAction rather than ActionEvent

    @property
    def card(self):
        return self.action.card

    def __str__(self):
        return f'{self.player} plays {self.action}'



import traceback

import numpy as np
import torch

from .utils import log
from rlcard.utils import run_game_pettingzoo

def create_buffers_pettingzoo(
    T,
    num_buffers,
    env,
    device_iterator,
):
    buffers = {}
    for device in device_iterator:
        buffers[device] = []
        for agent_name in env.agents:
            state_shape = env.observation_space(agent_name)["observation"].shape
            specs = dict(
                done=dict(size=(T,), dtype=torch.bool),
                episode_return=dict(size=(T,), dtype=torch.float32),
                target=dict(size=(T,), dtype=torch.float32),
                state=dict(size=(T,)+tuple(state_shape), dtype=torch.int8),
                action=dict(size=(T,)+(env.action_space(agent_name).n,), dtype=torch.int8),
            )
            _buffers = {key: [] for key in specs}
            for _ in range(num_buffers):
                for key in _buffers:
                    if device == "cpu":
                        _buffer = torch.empty(**specs[key]).to('cpu').share_memory_()
                    else:
                        _buffer = torch.empty(**specs[key]).to('cuda:'+str(device)).share_memory_()
                    _buffers[key].append(_buffer)
            buffers[device].append(_buffers)
    return buffers

def _get_action_feature(action, action_space):
    out = np.zeros(action_space)
    out[action] = 1
    return out

def act_pettingzoo(
    i,
    device,
    T,
    free_queue,
    full_queue,
    model,
    buffers,
    env
):
    log.info('Device %s Actor %i started.', str(device), i)
    try:
        done_buf = [[] for _ in range(env.num_agents)]
        episode_return_buf = [[] for _ in range(env.num_agents)]
        target_buf = [[] for _ in range(env.num_agents)]
        state_buf = [[] for _ in range(env.num_agents)]
        action_buf = [[] for _ in range(env.num_agents)]
        size = [0 for _ in range(env.num_agents)]

        while True:
            trajectories = run_game_pettingzoo(env, model.agents, is_training=True)
            for agent_id, agent_name in enumerate(env.possible_agents):
                traj_size = len(trajectories[agent_name]) // 2
                if traj_size > 0:
                    size[agent_id] += traj_size
                    target_return = trajectories[agent_name][-2][1]
                    target_buf[agent_id].extend([target_return for _ in range(traj_size)])
                    for i in range(0, len(trajectories[agent_name]), 2):
                        state = trajectories[agent_name][i][0]['observation']
                        action = _get_action_feature(
                            trajectories[agent_name][i+1], model.agents[agent_name].action_shape
                        )
                        episode_return = trajectories[agent_name][i][1]
                        done = trajectories[agent_name][i][2]
                        state_buf[agent_id].append(torch.from_numpy(state))
                        action_buf[agent_id].append(torch.from_numpy(action))
                        episode_return_buf[agent_id].append(episode_return)
                        done_buf[agent_id].append(done)

                while size[agent_id] > T:
                    index = free_queue[agent_id].get()
                    if index is None:
                        print("index is None")
                        break
                    for t in range(T):
                        temp_done = done_buf[agent_id][t]
                        buffers[agent_id]['done'][index][t, ...] = temp_done
                        buffers[agent_id]['episode_return'][index][t, ...] = episode_return_buf[agent_id][t]
                        buffers[agent_id]['target'][index][t, ...] = target_buf[agent_id][t]
                        buffers[agent_id]['state'][index][t, ...] = state_buf[agent_id][t]
                        buffers[agent_id]['action'][index][t, ...] = action_buf[agent_id][t]
                    full_queue[agent_id].put(index)
                    done_buf[agent_id] = done_buf[agent_id][T:]
                    episode_return_buf[agent_id] = episode_return_buf[agent_id][T:]
                    target_buf[agent_id] = target_buf[agent_id][T:]
                    state_buf[agent_id] = state_buf[agent_id][T:]
                    action_buf[agent_id] = action_buf[agent_id][T:]
                    size[agent_id] -= T

    except KeyboardInterrupt:
        pass
    except Exception as e:
        log.error('Exception in worker process %i', i)
        traceback.print_exc()
        raise e



'''
    Project: Gui Gin Rummy
    File name: handling_tap_player_pane.py
    Author: William Hale
    Date created: 3/14/2020
'''

# from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .game_canvas import GameCanvas

from . import configurations
from . import utils

from .canvas_item import CanvasItem


def handle_tap_player_pane(hit_item: CanvasItem, event, game_canvas: 'GameCanvas'):
    # un-select and un-jog all held cards
    player_id = None
    if game_canvas.player_panes[0] == hit_item:
        player_id = 0
    elif game_canvas.player_panes[1] == hit_item:
        player_id = 1
    if player_id is not None and game_canvas.query.is_human(player_id):
        held_pile_item_ids = game_canvas.getter.get_held_pile_item_ids(player_id)
        for item_id in held_pile_item_ids:
            game_canvas.dtag(item_id, configurations.JOGGED_TAG)
            game_canvas.dtag(item_id, configurations.SELECTED_TAG)
        utils.fan_held_pile(player_id=player_id, game_canvas=game_canvas)



from rlcard.games.base import Card
from rlcard.games.limitholdem import Dealer

class LeducholdemDealer(Dealer):

    def __init__(self, np_random):
        ''' Initialize a leducholdem dealer class
        '''
        self.np_random = np_random
        self.deck = [Card('S', 'J'), Card('H', 'J'), Card('S', 'Q'), Card('H', 'Q'), Card('S', 'K'), Card('H', 'K')]
        self.shuffle()
        self.pot = 0




class BlackjackJudger:
    def __init__(self, np_random):
        ''' Initialize a BlackJack judger class
        '''
        self.np_random = np_random
        self.rank2score = {"A":11, "2":2, "3":3, "4":4, "5":5, "6":6, "7":7, "8":8, "9":9, "T":10, "J":10, "Q":10, "K":10}

    def judge_round(self, player):
        ''' Judge the target player's status

        Args:
            player (int): target player's id

        Returns:
            status (str): the status of the target player
            score (int): the current score of the player
        '''
        score = self.judge_score(player.hand)
        if score <= 21:
            return "alive", score
        else:
            return "bust", score

    def judge_game(self, game, game_pointer):
        ''' Judge the winner of the game

        Args:
            game (class): target game class
        '''
        '''
                game.winner['dealer'] doesn't need anymore if we change code like this

                player bust (whether dealer bust or not) => game.winner[playerX] = -1
                player and dealer tie => game.winner[playerX] = 1
                dealer bust and player not bust => game.winner[playerX] = 2
                player get higher score than dealer => game.winner[playerX] = 2
                dealer get higher score than player => game.winner[playerX] = -1
                game.winner[playerX] = 0 => the game is still ongoing
                '''

        if game.players[game_pointer].status == 'bust':
            game.winner['player' + str(game_pointer)] = -1
        elif game.dealer.status == 'bust':
            game.winner['player' + str(game_pointer)] = 2
        else:
            if game.players[game_pointer].score > game.dealer.score:
                game.winner['player' + str(game_pointer)] = 2
            elif game.players[game_pointer].score < game.dealer.score:
                game.winner['player' + str(game_pointer)] = -1
            else:
                game.winner['player' + str(game_pointer)] = 1

    def judge_score(self, cards):
        ''' Judge the score of a given cards set

        Args:
            cards (list): a list of cards

        Returns:
            score (int): the score of the given cards set
        '''
        score = 0
        count_a = 0
        for card in cards:
            card_score = self.rank2score[card.rank]
            score += card_score
            if card.rank == 'A':
                count_a += 1
        while score > 21 and count_a > 0:
            count_a -= 1
            score -= 10
        return score



#The MIT License
#
#Copyright (c) 2020 DATA Lab at Texas A&M University
#Copyright (c) 2016 OpenAI (https://openai.com)
#
#Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import hashlib
import numpy as np
import os
import struct

def colorize(string, color, bold=False, highlight = False):
    """Return string surrounded by appropriate terminal color codes to
    print colorized text.  Valid colors: gray, red, green, yellow,
    blue, magenta, cyan, white, crimson
    """
    attr = []
    num = color2num[color]
    if highlight: num += 10
    attr.append(str(num))
    if bold: attr.append('1')
    attrs = ';'.join(attr)
    return '\x1b[%sm%s\x1b[0m' % (attrs, string)

def error(msg, *args):
    print(colorize('%s: %s'%('ERROR', msg % args), 'red'))

def np_random(seed=None):
    if seed is not None and not (isinstance(seed, int) and 0 <= seed):
        raise error.Error('Seed must be a non-negative integer or omitted, not {}'.format(seed))

    seed = create_seed(seed)

    rng = np.random.RandomState()
    rng.seed(_int_list_from_bigint(hash_seed(seed)))
    return rng, seed

def hash_seed(seed=None, max_bytes=8):
    """Any given evaluation is likely to have many PRNG's active at
    once. (Most commonly, because the environment is running in
    multiple processes.) There's literature indicating that having
    linear correlations between seeds of multiple PRNG's can correlate
    the outputs:

    http://blogs.unity3d.com/2015/01/07/a-primer-on-repeatable-random-numbers/
    http://stackoverflow.com/questions/1554958/how-different-do-random-seeds-need-to-be
    http://dl.acm.org/citation.cfm?id=1276928

    Thus, for sanity we hash the seeds before using them. (This scheme
    is likely not crypto-strength, but it should be good enough to get
    rid of simple correlations.)

    Args:
        seed (Optional[int]): None seeds from an operating system specific randomness source.
        max_bytes: Maximum number of bytes to use in the hashed seed.
    """
    if seed is None:
        seed = create_seed(max_bytes=max_bytes)
    _hash = hashlib.sha512(str(seed).encode('utf8')).digest()
    return _bigint_from_bytes(_hash[:max_bytes])

def create_seed(a=None, max_bytes=8):
    """Create a strong random seed. Otherwise, Python 2 would seed using
    the system time, which might be non-robust especially in the
    presence of concurrency.

    Args:
        a (Optional[int, str]): None seeds from an operating system specific randomness source.
        max_bytes: Maximum number of bytes to use in the seed.
    """
    # Adapted from https://svn.python.org/projects/python/tags/r32/Lib/random.py
    if a is None:
        a = _bigint_from_bytes(os.urandom(max_bytes))
    elif isinstance(a, str):
        a = a.encode('utf8')
        a += hashlib.sha512(a).digest()
        a = _bigint_from_bytes(a[:max_bytes])
    elif isinstance(a, int):
        a = a % 2**(8 * max_bytes)
    else:
        raise error.Error('Invalid type for seed: {} ({})'.format(type(a), a))

    return a

# TODO: don't hardcode sizeof_int here
def _bigint_from_bytes(_bytes):
    sizeof_int = 4
    padding = sizeof_int - len(_bytes) % sizeof_int
    _bytes += b'\0' * padding
    int_count = int(len(_bytes) / sizeof_int)
    unpacked = struct.unpack("{}I".format(int_count), _bytes)
    accum = 0
    for i, val in enumerate(unpacked):
        accum += 2 ** (sizeof_int * 8 * i) * val
    return accum

def _int_list_from_bigint(bigint):
    # Special case 0
    if bigint < 0:
        raise error.Error('Seed must be non-negative, not {}'.format(bigint))
    elif bigint == 0:
        return [0]

    ints = []
    while bigint > 0:
        bigint, mod = divmod(bigint, 2 ** 32)
        ints.append(mod)
    return ints



import unittest
import numpy as np

import rlcard
from rlcard.agents.random_agent import RandomAgent
from .determism_util import is_deterministic

class TestMahjongEnv(unittest.TestCase):

    def test_reset_and_extract_state(self):
        env = rlcard.make('mahjong')
        state, _ = env.reset()
        self.assertEqual(state['obs'].size, 816)

    def test_is_deterministic(self):
        self.assertTrue(is_deterministic('mahjong'))

    def test_get_legal_actions(self):
        env = rlcard.make('mahjong')
        env.set_agents([RandomAgent(env.num_actions) for _ in range(env.num_players)])
        env.reset()
        legal_actions = env._get_legal_actions()
        for legal_action in legal_actions:
            self.assertLessEqual(legal_action, env.num_actions-1)

    def test_step(self):
        env = rlcard.make('mahjong')
        state, _ = env.reset()
        action = np.random.choice(list(state['legal_actions'].keys()))
        _, player_id = env.step(action)
        self.assertEqual(player_id, env.game.round.current_player)

    def test_step_back(self):
        env = rlcard.make('mahjong', config={'allow_step_back':True})
        state, player_id = env.reset()
        action = np.random.choice(list(state['legal_actions'].keys()))
        env.step(action)
        env.step_back()
        self.assertEqual(env.game.round.current_player, player_id)

        env = rlcard.make('mahjong', config={'allow_step_back':False})
        state, player_id = env.reset()
        action = np.random.choice(list(state['legal_actions'].keys()))
        env.step(action)
        # env.step_back()
        self.assertRaises(Exception, env.step_back)

    def test_run(self):
        env = rlcard.make('mahjong')
        env.set_agents([RandomAgent(env.num_actions) for _ in range(env.num_players)])
        trajectories, payoffs = env.run(is_training=False)
        trajectories, payoffs = env.run(is_training=True)

if __name__ == '__main__':
    unittest.main()



import unittest
import torch
import numpy as np

from rlcard.agents.dqn_agent import DQNAgent

class TestDQN(unittest.TestCase):

    def test_init(self):

        agent = DQNAgent(replay_memory_size=0,
                         replay_memory_init_size=0,
                         update_target_estimator_every=0,
                         discount_factor=0,
                         epsilon_start=0,
                         epsilon_end=0,
                         epsilon_decay_steps=0,
                         batch_size=0,
                         num_actions=2,
                         state_shape=[1],
                         mlp_layers=[10,10],
                         device=torch.device('cpu'))

        self.assertEqual(agent.replay_memory_init_size, 0)
        self.assertEqual(agent.update_target_estimator_every, 0)
        self.assertEqual(agent.discount_factor, 0)
        self.assertEqual(agent.epsilon_decay_steps, 0)
        self.assertEqual(agent.batch_size, 0)
        self.assertEqual(agent.num_actions, 2)

    def test_train(self):

        memory_init_size = 100
        num_steps = 500

        agent = DQNAgent(replay_memory_size = 200,
                         replay_memory_init_size=memory_init_size,
                         update_target_estimator_every=100,
                         state_shape=[2],
                         mlp_layers=[10,10],
                         device=torch.device('cpu'))

        predicted_action, _ = agent.eval_step({'obs': np.random.random_sample((2,)), 'legal_actions': {0: None, 1: None}, 'raw_legal_actions': ['call', 'raise']})
        self.assertGreaterEqual(predicted_action, 0)
        self.assertLessEqual(predicted_action, 1)

        for _ in range(num_steps):
            ts = [{'obs': np.random.random_sample((2,)), 'legal_actions': {0: None, 1: None}}, np.random.randint(2), 0, {'obs': np.random.random_sample((2,)), 'legal_actions': {0: None, 1: None}, 'raw_legal_actions': ['call', 'raise']}, True]
            agent.feed(ts)

        predicted_action = agent.step({'obs': np.random.random_sample((2,)), 'legal_actions': {0: None, 1: None}})
        self.assertGreaterEqual(predicted_action, 0)
        self.assertLessEqual(predicted_action, 1)



''' Register rule-based models or pre-trianed models
'''
from rlcard.models.registration import register, load

register(
    model_id = 'leduc-holdem-cfr',
    entry_point='rlcard.models.pretrained_models:LeducHoldemCFRModel')

register(
    model_id = 'leduc-holdem-rule-v1',
    entry_point='rlcard.models.leducholdem_rule_models:LeducHoldemRuleModelV1')

register(
    model_id = 'leduc-holdem-rule-v2',
    entry_point='rlcard.models.leducholdem_rule_models:LeducHoldemRuleModelV2')

register(
    model_id = 'uno-rule-v1',
    entry_point='rlcard.models.uno_rule_models:UNORuleModelV1')

register(
    model_id = 'limit-holdem-rule-v1',
    entry_point='rlcard.models.limitholdem_rule_models:LimitholdemRuleModelV1')

register(
    model_id = 'doudizhu-rule-v1',
    entry_point='rlcard.models.doudizhu_rule_models:DouDizhuRuleModelV1')

register(
    model_id='gin-rummy-novice-rule',
    entry_point='rlcard.models.gin_rummy_rule_models:GinRummyNoviceRuleModel')



'''
    File name: envs/bridge.py
    Author: William Hale
    Date created: 11/26/2021
'''

import numpy as np
from collections import OrderedDict

from rlcard.envs import Env

from rlcard.games.bridge import Game

from rlcard.games.bridge.game import BridgeGame
from rlcard.games.bridge.utils.action_event import ActionEvent
from rlcard.games.bridge.utils.bridge_card import BridgeCard
from rlcard.games.bridge.utils.move import CallMove, PlayCardMove

#   [] Why no_bid_action_id in bidding_rep ?
#       It allows the bidding always to start with North.
#       If North is not the dealer, then he must call 'no_bid'.
#       Until the dealer is reached, 'no_bid' must be the call.
#       I think this might help because it keeps a player's bid in a fixed 'column'.
#       Note: the 'no_bid' is only inserted in the bidding_rep, not in the actual game.
#
#   [] Why current_player_rep ?
#       Explanation here.
#
#   [] Note: hands_rep maintain the hands by N, E, S, W.
#
#   [] Note: trick_rep maintains the trick cards by N, E, S, W.
#      The trick leader can be deduced since play is in clockwise direction.
#
#   [] Note: is_bidding_rep can be deduced from bidding_rep.
#      I think I added is_bidding_rep before bidding_rep and thus it helped in early testing.
#      My early testing had just the player's hand: I think the model conflated the bidding phase with the playing phase in this situation.
#      Although is_bidding_rep is not needed, keeping it may improve learning.
#
#   [] Note: bidding_rep uses the action_id instead of one hot encoding.
#      I think one hot encoding would make the input dimension significantly larger.
#


class BridgeEnv(Env):
    ''' Bridge Environment
    '''
    def __init__(self, config):
        self.name = 'bridge'
        self.game = Game()
        super().__init__(config=config)
        self.bridgePayoffDelegate = DefaultBridgePayoffDelegate()
        self.bridgeStateExtractor = DefaultBridgeStateExtractor()
        state_shape_size = self.bridgeStateExtractor.get_state_shape_size()
        self.state_shape = [[1, state_shape_size] for _ in range(self.num_players)]
        self.action_shape = [None for _ in range(self.num_players)]

    def get_payoffs(self):
        ''' Get the payoffs of players.

        Returns:
            (list): A list of payoffs for each player.
        '''
        return self.bridgePayoffDelegate.get_payoffs(game=self.game)

    def get_perfect_information(self):
        ''' Get the perfect information of the current state

        Returns:
            (dict): A dictionary of all the perfect information of the current state
        '''
        return self.game.round.get_perfect_information()

    def _extract_state(self, state):  # wch: don't use state 211126
        ''' Extract useful information from state for RL.

        Args:
            state (dict): The raw state

        Returns:
            (numpy.array): The extracted state
        '''
        return self.bridgeStateExtractor.extract_state(game=self.game)

    def _decode_action(self, action_id):
        ''' Decode Action id to the action in the game.

        Args:
            action_id (int): The id of the action

        Returns:
            (ActionEvent): The action that will be passed to the game engine.
        '''
        return ActionEvent.from_action_id(action_id=action_id)

    def _get_legal_actions(self):
        ''' Get all legal actions for current state.

        Returns:
            (list): A list of legal actions' id.
        '''
        raise NotImplementedError  # wch: not needed


class BridgePayoffDelegate(object):

    def get_payoffs(self, game: BridgeGame):
        ''' Get the payoffs of players. Must be implemented in the child class.

        Returns:
            (list): A list of payoffs for each player.

        Note: Must be implemented in the child class.
        '''
        raise NotImplementedError


class DefaultBridgePayoffDelegate(BridgePayoffDelegate):

    def __init__(self):
        self.make_bid_bonus = 2

    def get_payoffs(self, game: BridgeGame):
        ''' Get the payoffs of players.

        Returns:
            (list): A list of payoffs for each player.
        '''
        contract_bid_move = game.round.contract_bid_move
        if contract_bid_move:
            declarer = contract_bid_move.player
            bid_trick_count = contract_bid_move.action.bid_amount + 6
            won_trick_counts = game.round.won_trick_counts
            declarer_won_trick_count = won_trick_counts[declarer.player_id % 2]
            defender_won_trick_count = won_trick_counts[(declarer.player_id + 1) % 2]
            declarer_payoff = bid_trick_count + self.make_bid_bonus if bid_trick_count <= declarer_won_trick_count else declarer_won_trick_count - bid_trick_count
            defender_payoff = defender_won_trick_count
            payoffs = []
            for player_id in range(4):
                payoff = declarer_payoff if player_id % 2 == declarer.player_id % 2 else defender_payoff
                payoffs.append(payoff)
        else:
            payoffs = [0, 0, 0, 0]
        return np.array(payoffs)


class BridgeStateExtractor(object):  # interface

    def get_state_shape_size(self) -> int:
        raise NotImplementedError

    def extract_state(self, game: BridgeGame):
        ''' Extract useful information from state for RL. Must be implemented in the child class.

        Args:
            game (BridgeGame): The game

        Returns:
            (numpy.array): The extracted state
        '''
        raise NotImplementedError

    @staticmethod
    def get_legal_actions(game: BridgeGame):
        ''' Get all legal actions for current state.

        Returns:
            (OrderedDict): A OrderedDict of legal actions' id.
        '''
        legal_actions = game.judger.get_legal_actions()
        legal_actions_ids = {action_event.action_id: None for action_event in legal_actions}
        return OrderedDict(legal_actions_ids)


class DefaultBridgeStateExtractor(BridgeStateExtractor):

    def __init__(self):
        super().__init__()
        self.max_bidding_rep_index = 40  # Note: max of 40 calls
        self.last_bid_rep_size = 1 + 35 + 3  # no_bid, bid, pass, dbl, rdbl

    def get_state_shape_size(self) -> int:
        state_shape_size = 0
        state_shape_size += 4 * 52  # hands_rep_size
        state_shape_size += 4 * 52  # trick_rep_size
        state_shape_size += 52  # hidden_cards_rep_size
        state_shape_size += 4  # vul_rep_size
        state_shape_size += 4  # dealer_rep_size
        state_shape_size += 4  # current_player_rep_size
        state_shape_size += 1  # is_bidding_rep_size
        state_shape_size += self.max_bidding_rep_index  # bidding_rep_size
        state_shape_size += self.last_bid_rep_size  # last_bid_rep_size
        state_shape_size += 8  # bid_amount_rep_size
        state_shape_size += 5  # trump_suit_rep_size
        return state_shape_size

    def extract_state(self, game: BridgeGame):
        ''' Extract useful information from state for RL.

        Args:
            game (BridgeGame): The game

        Returns:
            (numpy.array): The extracted state
        '''
        extracted_state = {}
        legal_actions: OrderedDict = self.get_legal_actions(game=game)
        raw_legal_actions = list(legal_actions.keys())
        current_player = game.round.get_current_player()
        current_player_id = current_player.player_id

        # construct hands_rep of hands of players
        hands_rep = [np.zeros(52, dtype=int) for _ in range(4)]
        if not game.is_over():
            for card in game.round.players[current_player_id].hand:
                hands_rep[current_player_id][card.card_id] = 1
            if game.round.is_bidding_over():
                dummy = game.round.get_dummy()
                other_known_player = dummy if dummy.player_id != current_player_id else game.round.get_declarer()
                for card in other_known_player.hand:
                    hands_rep[other_known_player.player_id][card.card_id] = 1

        # construct trick_pile_rep
        trick_pile_rep = [np.zeros(52, dtype=int) for _ in range(4)]
        if game.round.is_bidding_over() and not game.is_over():
            trick_moves = game.round.get_trick_moves()
            for move in trick_moves:
                player = move.player
                card = move.card
                trick_pile_rep[player.player_id][card.card_id] = 1

        # construct hidden_card_rep (during trick taking phase)
        hidden_cards_rep = np.zeros(52, dtype=int)
        if not game.is_over():
            if game.round.is_bidding_over():
                declarer = game.round.get_declarer()
                if current_player_id % 2 == declarer.player_id % 2:
                    hidden_player_ids = [(current_player_id + 1) % 4, (current_player_id + 3) % 4]
                else:
                    hidden_player_ids = [declarer.player_id, (current_player_id + 2) % 4]
                for hidden_player_id in hidden_player_ids:
                    for card in game.round.players[hidden_player_id].hand:
                        hidden_cards_rep[card.card_id] = 1
            else:
                for player in game.round.players:
                    if player.player_id != current_player_id:
                        for card in player.hand:
                            hidden_cards_rep[card.card_id] = 1

        # construct vul_rep
        vul_rep = np.array(game.round.tray.vul, dtype=int)

        # construct dealer_rep
        dealer_rep = np.zeros(4, dtype=int)
        dealer_rep[game.round.tray.dealer_id] = 1

        # construct current_player_rep
        current_player_rep = np.zeros(4, dtype=int)
        current_player_rep[current_player_id] = 1

        # construct is_bidding_rep
        is_bidding_rep = np.array([1] if game.round.is_bidding_over() else [0])

        # construct bidding_rep
        bidding_rep = np.zeros(self.max_bidding_rep_index, dtype=int)
        bidding_rep_index = game.round.dealer_id  # no_bid_action_ids allocated at start so that north always 'starts' the bidding
        for move in game.round.move_sheet:
            if bidding_rep_index >= self.max_bidding_rep_index:
                break
            elif isinstance(move, PlayCardMove):
                break
            elif isinstance(move, CallMove):
                bidding_rep[bidding_rep_index] = move.action.action_id
                bidding_rep_index += 1

        # last_bid_rep
        last_bid_rep = np.zeros(self.last_bid_rep_size, dtype=int)
        last_move = game.round.move_sheet[-1]
        if isinstance(last_move, CallMove):
            last_bid_rep[last_move.action.action_id - ActionEvent.no_bid_action_id] = 1

        # bid_amount_rep and trump_suit_rep
        bid_amount_rep = np.zeros(8, dtype=int)
        trump_suit_rep = np.zeros(5, dtype=int)
        if game.round.is_bidding_over() and not game.is_over() and game.round.play_card_count == 0:
            contract_bid_move = game.round.contract_bid_move
            if contract_bid_move:
                bid_amount_rep[contract_bid_move.action.bid_amount] = 1
                bid_suit = contract_bid_move.action.bid_suit
                bid_suit_index = 4 if not bid_suit else BridgeCard.suits.index(bid_suit)
                trump_suit_rep[bid_suit_index] = 1

        rep = []
        rep += hands_rep
        rep += trick_pile_rep
        rep.append(hidden_cards_rep)
        rep.append(vul_rep)
        rep.append(dealer_rep)
        rep.append(current_player_rep)
        rep.append(is_bidding_rep)
        rep.append(bidding_rep)
        rep.append(last_bid_rep)
        rep.append(bid_amount_rep)
        rep.append(trump_suit_rep)

        obs = np.concatenate(rep)
        extracted_state['obs'] = obs
        extracted_state['legal_actions'] = legal_actions
        extracted_state['raw_legal_actions'] = raw_legal_actions
        extracted_state['raw_obs'] = obs
        return extracted_state



import unittest
import numpy as np

import rlcard
from rlcard.agents.random_agent import RandomAgent
from .determism_util import is_deterministic

class TestBlackjackEnv(unittest.TestCase):

    def test_init_and_extract_state(self):
        env = rlcard.make('blackjack')
        state, _ = env.reset()
        for score in state['obs']:
            self.assertLessEqual(score, 30)

    def test_is_deterministic(self):
        self.assertTrue(is_deterministic('blackjack'))

    def test_decode_action(self):
        env = rlcard.make('blackjack')
        self.assertEqual(env._decode_action(0), 'hit')
        self.assertEqual(env._decode_action(1), 'stand')

    def test_get_legal_actions(self):
        env = rlcard.make('blackjack')
        actions = env._get_legal_actions()
        self.assertEqual(len(actions), 2)
        self.assertEqual(actions[0], 0)
        self.assertEqual(actions[1], 1)

    def test_get_payoffs(self):
        env = rlcard.make('blackjack')
        for _ in range(100):
            env.reset()
            while not env.is_over():
                action = np.random.choice([0, 1])
                env.step(action)
            payoffs = env.get_payoffs()
            for payoff in payoffs:
                self.assertIn(payoff, [-1, 1, 0])

    def test_step_back(self):
        env = rlcard.make('blackjack', config={'allow_step_back':True})
        _, player_id = env.reset()
        env.step(1)
        _, back_player_id = env.step_back()
        self.assertEqual(player_id, back_player_id)
        self.assertEqual(env.step_back(), False)

        env = rlcard.make('blackjack')
        with self.assertRaises(Exception):
            env.step_back()

    def test_multiplayers(self):
        env = rlcard.make('blackjack', config={'game_num_players':5})
        num_players = env.game.get_num_players()
        self.assertEqual(num_players, 5)

    def test_run(self):
        env = rlcard.make('blackjack')
        env.set_agents([RandomAgent(env.num_actions)])
        trajectories, _ = env.run(is_training=False)
        self.assertEqual(len(trajectories), 1)
        trajectories, _ = env.run(is_training=True)
        self.assertEqual(len(trajectories), 1)

if __name__ == '__main__':
    unittest.main()



'''
    File name: gin_rummy/round.py
    Author: William Hale
    Date created: 2/12/2020
'''
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .utils.move import GinRummyMove

from typing import List

from rlcard.games.gin_rummy.dealer import GinRummyDealer

from .utils.action_event import DrawCardAction, PickUpDiscardAction, DeclareDeadHandAction
from .utils.action_event import DiscardAction, KnockAction, GinAction
from .utils.action_event import ScoreNorthPlayerAction, ScoreSouthPlayerAction

from .utils.move import DealHandMove
from .utils.move import DrawCardMove, PickupDiscardMove, DeclareDeadHandMove
from .utils.move import DiscardMove, KnockMove, GinMove
from .utils.move import ScoreNorthMove, ScoreSouthMove

from .utils.gin_rummy_error import GinRummyProgramError

from .player import GinRummyPlayer
from . import judge

from rlcard.games.gin_rummy.utils import melding
from rlcard.games.gin_rummy.utils import utils


class GinRummyRound:

    def __init__(self, dealer_id: int, np_random):
        ''' Initialize the round class

            The round class maintains the following instances:
                1) dealer: the dealer of the round; dealer has stock_pile and discard_pile
                2) players: the players in the round; each player has his own hand_pile
                3) current_player_id: the id of the current player who has the move
                4) is_over: true if the round is over
                5) going_out_action: knock or gin or None
                6) going_out_player_id: id of player who went out or None
                7) move_sheet: history of the moves of the player (including the deal_hand_move)

            The round class maintains a list of moves made by the players in self.move_sheet.
            move_sheet is similar to a chess score sheet.
            I didn't want to call it a score_sheet since it is not keeping score.
            I could have called move_sheet just moves, but that might conflict with the name moves used elsewhere.
            I settled on the longer name "move_sheet" to indicate that it is the official list of moves being made.

        Args:
            dealer_id: int
        '''
        self.np_random = np_random
        self.dealer_id = dealer_id
        self.dealer = GinRummyDealer(self.np_random)
        self.players = [GinRummyPlayer(player_id=0, np_random=self.np_random), GinRummyPlayer(player_id=1, np_random=self.np_random)]
        self.current_player_id = (dealer_id + 1) % 2
        self.is_over = False
        self.going_out_action = None  # going_out_action: int or None
        self.going_out_player_id = None  # going_out_player_id: int or None
        self.move_sheet = []  # type: List[GinRummyMove]
        player_dealing = GinRummyPlayer(player_id=dealer_id, np_random=self.np_random)
        shuffled_deck = self.dealer.shuffled_deck
        self.move_sheet.append(DealHandMove(player_dealing=player_dealing, shuffled_deck=shuffled_deck))

    def get_current_player(self) -> GinRummyPlayer or None:
        current_player_id = self.current_player_id
        return None if current_player_id is None else self.players[current_player_id]

    def draw_card(self, action: DrawCardAction):
        # when current_player takes DrawCardAction step, the move is recorded and executed
        # current_player keeps turn
        current_player = self.players[self.current_player_id]
        if not len(current_player.hand) == 10:
            raise GinRummyProgramError("len(current_player.hand) is {}: should be 10.".format(len(current_player.hand)))
        card = self.dealer.stock_pile.pop()
        self.move_sheet.append(DrawCardMove(current_player, action=action, card=card))
        current_player.add_card_to_hand(card=card)

    def pick_up_discard(self, action: PickUpDiscardAction):
        # when current_player takes PickUpDiscardAction step, the move is recorded and executed
        # opponent knows that the card is in current_player hand
        # current_player keeps turn
        current_player = self.players[self.current_player_id]
        if not len(current_player.hand) == 10:
            raise GinRummyProgramError("len(current_player.hand) is {}: should be 10.".format(len(current_player.hand)))
        card = self.dealer.discard_pile.pop()
        self.move_sheet.append(PickupDiscardMove(current_player, action, card=card))
        current_player.add_card_to_hand(card=card)
        current_player.known_cards.append(card)

    def declare_dead_hand(self, action: DeclareDeadHandAction):
        # when current_player takes DeclareDeadHandAction step, the move is recorded and executed
        # north becomes current_player to score his hand
        current_player = self.players[self.current_player_id]
        self.move_sheet.append(DeclareDeadHandMove(current_player, action))
        self.going_out_action = action
        self.going_out_player_id = self.current_player_id
        if not len(current_player.hand) == 10:
            raise GinRummyProgramError("len(current_player.hand) is {}: should be 10.".format(len(current_player.hand)))
        self.current_player_id = 0

    def discard(self, action: DiscardAction):
        # when current_player takes DiscardAction step, the move is recorded and executed
        # opponent knows that the card is no longer in current_player hand
        # current_player loses his turn and the opponent becomes the current player
        current_player = self.players[self.current_player_id]
        if not len(current_player.hand) == 11:
            raise GinRummyProgramError("len(current_player.hand) is {}: should be 11.".format(len(current_player.hand)))
        self.move_sheet.append(DiscardMove(current_player, action))
        card = action.card
        current_player.remove_card_from_hand(card=card)
        if card in current_player.known_cards:
            current_player.known_cards.remove(card)
        self.dealer.discard_pile.append(card)
        self.current_player_id = (self.current_player_id + 1) % 2

    def knock(self, action: KnockAction):
        # when current_player takes KnockAction step, the move is recorded and executed
        # opponent knows that the card is no longer in current_player hand
        # north becomes current_player to score his hand
        current_player = self.players[self.current_player_id]
        self.move_sheet.append(KnockMove(current_player, action))
        self.going_out_action = action
        self.going_out_player_id = self.current_player_id
        if not len(current_player.hand) == 11:
            raise GinRummyProgramError("len(current_player.hand) is {}: should be 11.".format(len(current_player.hand)))
        card = action.card
        current_player.remove_card_from_hand(card=card)
        if card in current_player.known_cards:
            current_player.known_cards.remove(card)
        self.current_player_id = 0

    def gin(self, action: GinAction, going_out_deadwood_count: int):
        # when current_player takes GinAction step, the move is recorded and executed
        # opponent knows that the card is no longer in current_player hand
        # north becomes current_player to score his hand
        current_player = self.players[self.current_player_id]
        self.move_sheet.append(GinMove(current_player, action))
        self.going_out_action = action
        self.going_out_player_id = self.current_player_id
        if not len(current_player.hand) == 11:
            raise GinRummyProgramError("len(current_player.hand) is {}: should be 11.".format(len(current_player.hand)))
        _, gin_cards = judge.get_going_out_cards(current_player.hand, going_out_deadwood_count)
        card = gin_cards[0]
        current_player.remove_card_from_hand(card=card)
        if card in current_player.known_cards:
            current_player.known_cards.remove(card)
        self.current_player_id = 0

    def score_player_0(self, action: ScoreNorthPlayerAction):
        # when current_player takes ScoreNorthPlayerAction step, the move is recorded and executed
        # south becomes current player
        if not self.current_player_id == 0:
            raise GinRummyProgramError("current_player_id is {}: should be 0.".format(self.current_player_id))
        current_player = self.get_current_player()
        best_meld_clusters = melding.get_best_meld_clusters(hand=current_player.hand)
        best_meld_cluster = [] if not best_meld_clusters else best_meld_clusters[0]
        deadwood_count = utils.get_deadwood_count(hand=current_player.hand, meld_cluster=best_meld_cluster)
        self.move_sheet.append(ScoreNorthMove(player=current_player,
                                              action=action,
                                              best_meld_cluster=best_meld_cluster,
                                              deadwood_count=deadwood_count))
        self.current_player_id = 1

    def score_player_1(self, action: ScoreSouthPlayerAction):
        # when current_player takes ScoreSouthPlayerAction step, the move is recorded and executed
        # south remains current player
        # the round is over
        if not self.current_player_id == 1:
            raise GinRummyProgramError("current_player_id is {}: should be 1.".format(self.current_player_id))
        current_player = self.get_current_player()
        best_meld_clusters = melding.get_best_meld_clusters(hand=current_player.hand)
        best_meld_cluster = [] if not best_meld_clusters else best_meld_clusters[0]
        deadwood_count = utils.get_deadwood_count(hand=current_player.hand, meld_cluster=best_meld_cluster)
        self.move_sheet.append(ScoreSouthMove(player=current_player,
                                              action=action,
                                              best_meld_cluster=best_meld_cluster,
                                              deadwood_count=deadwood_count))
        self.is_over = True



import unittest

from rlcard.games.limitholdem.player import PlayerStatus
from rlcard.games.nolimitholdem.game import NolimitholdemGame as Game, Stage
import numpy as np
from rlcard.utils import seeding

from rlcard.games.nolimitholdem.round import Action


class TestNolimitholdemMethods(unittest.TestCase):

    def test_init_game(self):
        game = Game()
        state, player_id = game.init_game()
        test_id = game.get_player_id()
        self.assertEqual(test_id, player_id)

    def test_step(self):
        game = Game()

        # test call
        game.init_game()
        init_not_raise_num = game.round.not_raise_num
        game.step(Action.CHECK_CALL)
        step_not_raise_num = game.round.not_raise_num
        self.assertEqual(init_not_raise_num + 1, step_not_raise_num)

        # test fold
        _, player_id = game.init_game()
        game.step(Action.FOLD)

        self.assertEqual(PlayerStatus.FOLDED, game.players[player_id].status)

        # test check
        game.init_game()
        game.step(Action.CHECK_CALL)

    def test_bet_more_than_chips(self):
        game = Game()

        # test check
        game.init_game()
        player = game.players[0]
        in_chips = player.in_chips
        player.bet(50)
        self.assertEqual(50+in_chips, player.in_chips)

        player.bet(150)
        self.assertEqual(100, player.in_chips)

    def test_step_2(self):
        game = Game()

        # test check
        game.init_game()
        self.assertEqual(Stage.PREFLOP, game.stage)
        game.step(Action.CHECK_CALL)
        game.step(Action.RAISE_POT)
        game.step(Action.CHECK_CALL)

        self.assertEqual(Stage.FLOP, game.stage)
        game.step(Action.CHECK_CALL)
        game.step(Action.CHECK_CALL)

        self.assertEqual(Stage.TURN, game.stage)
        game.step(Action.CHECK_CALL)
        game.step(Action.CHECK_CALL)

        self.assertEqual(Stage.RIVER, game.stage)

    def test_step_3_players(self):
        game = Game(num_players=3)

        # test check
        _, first_player_id = game.init_game()
        self.assertEqual(Stage.PREFLOP, game.stage)
        game.step(Action.CHECK_CALL)
        game.step(Action.CHECK_CALL)
        game.step(Action.RAISE_POT)
        game.step(Action.FOLD)
        game.step(Action.CHECK_CALL)

        self.assertEqual(Stage.FLOP, game.stage)
        self.assertEqual((first_player_id - 2) % 3, game.round.game_pointer)
        game.step(Action.CHECK_CALL)
        game.step(Action.RAISE_POT)
        game.step(Action.CHECK_CALL)

        self.assertEqual(Stage.TURN, game.stage)
        self.assertEqual((first_player_id - 2) % 3, game.round.game_pointer)
        game.step(Action.CHECK_CALL)
        game.step(Action.CHECK_CALL)

        self.assertEqual(Stage.RIVER, game.stage)

    def test_auto_step(self):
        game = Game()

        game.init_game()
        self.assertEqual(Stage.PREFLOP, game.stage)
        game.step(Action.ALL_IN)
        game.step(Action.CHECK_CALL)

        self.assertEqual(Stage.RIVER, game.stage)

    def test_all_in(self):
        game = Game()

        _, player_id = game.init_game()
        game.step(Action.ALL_IN)
        step_raised = game.round.raised[player_id]
        self.assertEqual(100, step_raised)
        self.assertEqual(100, game.players[player_id].in_chips)
        self.assertEqual(0, game.players[player_id].remained_chips)

    def test_all_in_rounds(self):
        game = Game()

        game.init_game()
        game.step(Action.CHECK_CALL)
        game.step(Action.CHECK_CALL)
        self.assertEqual(game.round_counter, 1)

        game.step(Action.CHECK_CALL)
        game.step(Action.ALL_IN)
        self.assertListEqual([Action.FOLD, Action.CHECK_CALL], game.get_legal_actions())
        game.step(Action.CHECK_CALL)
        self.assertEqual(game.round_counter, 4)
        self.assertEqual(200, game.dealer.pot)

    def test_raise_pot(self):
        game = Game()

        _, player_id = game.init_game()
        game.step(Action.RAISE_POT)
        step_raised = game.round.raised[player_id]
        self.assertEqual(4, step_raised)

        player_id = game.round.game_pointer
        game.step(Action.RAISE_POT)
        step_raised = game.round.raised[player_id]
        self.assertEqual(8, step_raised)

        player_id = game.round.game_pointer
        game.step(Action.RAISE_POT)
        step_raised = game.round.raised[player_id]
        self.assertEqual(16, step_raised)

        game.step(Action.CHECK_CALL)
        player_id = game.round.game_pointer
        game.step(Action.RAISE_POT)
        step_raised = game.round.raised[player_id]
        self.assertEqual(32, step_raised)

    def test_raise_half_pot(self):
        game = Game()

        _, player_id = game.init_game()
        self.assertNotIn(Action.RAISE_HALF_POT, game.get_legal_actions()) # Half pot equals call
        game.step(Action.CHECK_CALL)
        step_raised = game.round.raised[player_id]
        self.assertEqual(2, step_raised)

        player_id = game.round.game_pointer
        game.step(Action.RAISE_HALF_POT)
        step_raised = game.round.raised[player_id]
        self.assertEqual(4, step_raised)

        player_id = game.round.game_pointer
        game.step(Action.RAISE_HALF_POT)
        step_raised = game.round.raised[player_id]
        self.assertEqual(5, step_raised)

    def test_payoffs_1(self):
        game = Game()
        game.init_game()
        game.step(Action.CHECK_CALL)
        game.step(Action.RAISE_HALF_POT)
        game.step(Action.FOLD)
        self.assertTrue(game.is_over())
        self.assertEqual(2, len(game.get_payoffs()))
        #self.assertListEqual([-2.0, 2.0], game.get_payoffs())

    def test_payoffs_2(self):
        game = Game()
        np.random.seed(0)
        game.init_game()
        game.step(Action.CHECK_CALL)
        game.step(Action.RAISE_POT)
        game.step(Action.ALL_IN)
        game.step(Action.FOLD)
        self.assertTrue(game.is_over())
        self.assertEqual(2, len(game.get_payoffs()))
        #self.assertListEqual([6.0, -6.0], game.get_payoffs())

    def test_all_in_to_call(self):
        game = Game()
        game.init_chips = [50, 100]
        game.dealer_id = 0
        game.init_game()
        game.step(Action.CHECK_CALL)
        game.step(Action.ALL_IN)
        game.step(Action.CHECK_CALL)
        self.assertTrue(game.is_over())


if __name__ == '__main__':
    unittest.main()



from rlcard.utils.utils import init_standard_deck


class LimitHoldemDealer:
    def __init__(self, np_random):
        self.np_random = np_random
        self.deck = init_standard_deck()
        self.shuffle()
        self.pot = 0

    def shuffle(self):
        self.np_random.shuffle(self.deck)

    def deal_card(self):
        """
        Deal one card from the deck

        Returns:
            (Card): The drawn card from the deck
        """
        return self.deck.pop()



'''
    Project: Gui Gin Rummy
    File name: game_canvas_updater.py
    Author: William Hale
    Date created: 3/14/2020
'''

# from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .game_canvas import GameCanvas
    from rlcard.envs.gin_rummy import GinRummyEnv
from ..gin_rummy_human_agent import HumanAgent

from typing import List

from rlcard.games.gin_rummy.utils.action_event import ActionEvent
from rlcard.games.gin_rummy.utils.move import GinRummyMove
from rlcard.games.gin_rummy.utils.move import PlayerMove
from rlcard.games.gin_rummy.utils.move import DealHandMove
from rlcard.games.gin_rummy.utils.move import DrawCardMove, PickupDiscardMove, DeclareDeadHandMove
from rlcard.games.gin_rummy.utils.move import DiscardMove, KnockMove, GinMove
from rlcard.games.gin_rummy.utils.move import ScoreNorthMove, ScoreSouthMove
from rlcard.games.gin_rummy.utils.gin_rummy_error import GinRummyProgramError

from . import configurations
from . import handling_tap_discard_pile
from . import handling_tap_held_pile
from . import starting_new_game
from . import status_messaging
from . import utils

from .player_type import PlayerType

import rlcard.games.gin_rummy.utils.utils as gin_rummy_utils


class GameCanvasUpdater(object):

    def __init__(self, game_canvas: 'GameCanvas'):
        self.game_canvas = game_canvas
        self.env_thread = None
        self.pending_human_action_ids = []  # type: List[int]
        self.busy_body_id = None  # type: int or None
        self.is_stopped = False

    @property
    def mark(self) -> int:  # convenience property to briefly get mark
        return self.env_thread.mark

    @property
    def moves(self) -> List[GinRummyMove]:  # convenience property to briefly get moves
        return self.env.game.round.move_sheet

    @property
    def env(self) -> 'GinRummyEnv':
        return self.env_thread.gin_rummy_env

    @property
    def human_agent(self) -> 'HumanAgent' or None:
        south_agent = self.env_thread.gin_rummy_env.agents[1]
        return south_agent if isinstance(south_agent, HumanAgent) else None

    def apply_canvas_updates(self):
        if not self.env_thread.is_stopped:
            self._advance_mark()
            delay_ms = 1
            self.game_canvas.after(delay_ms, func=self.apply_canvas_updates)
        else:
            self.is_stopped = True

    def did_perform_actions(self, actions: List[int]):
        for action_id in actions:
            if self.game_canvas.player_types[self.busy_body_id] == PlayerType.human_player:
                self.pending_human_action_ids.append(action_id)
            self.env_thread.mark += 1
        self.busy_body_id = None

    #
    #   Private methods
    #

    def _advance_mark(self):
        if self.env.game.round is None:
            return
        if self.game_canvas.query.is_game_over() and self.mark >= len(self.moves):
            return
        if self.busy_body_id is not None:
            return
        if self.human_agent and not self.human_agent.is_choosing_action_id:
            return
        if self.human_agent and self.human_agent.chosen_action_id is not None:
            return
        if self.env_thread.is_action_id_available():
            move = self.moves[self.mark]
            thinking_time_in_ms = 0  # type: int
            if isinstance(move, DealHandMove):
                if not self.mark == 0:
                    raise GinRummyProgramError("mark={} must be 0.".format(self.mark))
                self.busy_body_id = move.player_dealing.player_id
            elif isinstance(move, ScoreNorthMove) or isinstance(move, ScoreSouthMove):
                self.busy_body_id = move.player.player_id
            elif isinstance(move, PlayerMove):
                self.busy_body_id = move.player.player_id
                thinking_time_in_ms = 1000  # type: int  # simulate the computer thinking
            else:
                raise GinRummyProgramError("GameCanvasUpdater advance_mark: unknown move={move}")
            if self.mark > 0:
                if self.game_canvas.player_types[self.busy_body_id] == PlayerType.human_player:
                    raise GinRummyProgramError("busy_body_id={} must not be human player.".format(self.busy_body_id))
            if not self.busy_body_id == self.game_canvas.getter.get_current_player_id():
                raise GinRummyProgramError("busy_body_id={} must equal current_player_id={}".format(self.busy_body_id, self.game_canvas.getter.get_current_player_id()))
            self._show_prolog_messages_on_computer_turn()
            self.game_canvas.after(thinking_time_in_ms, self._advance_mark_for_computer_player)
            return
        if self.pending_human_action_ids:
            action_id = self.pending_human_action_ids.pop(0)  # pending_human_action_ids is queue
            if utils.is_debug():
                action_event = ActionEvent.decode_action(action_id=action_id)
                print("S {}".format(action_event))  # FIXME: South may not always be actor
            self.human_agent.chosen_action_id = action_id
            return
        if not self.mark >= len(self.moves):  # FIXME: should be no pending computer moves
            raise GinRummyProgramError("Should be no pending computer moves.")
        waiting_player_id = self.env_thread.get_waiting_player_id()
        if waiting_player_id is None:
            return
        # FIXME: should be no pending computer moves
        if self.human_agent.chosen_action_id is not None:
            raise GinRummyProgramError("self.human_agent.chosen_action_id must not be None.")
        if self.busy_body_id is not None:
            raise GinRummyProgramError("busy_body_id={} must be None.".format(self.busy_body_id))
        if not waiting_player_id == self.game_canvas.getter.get_current_player_id():
            raise GinRummyProgramError("waiting_player_id={} must be current_player_id.".format(waiting_player_id))
        self.busy_body_id = waiting_player_id
        if not self.game_canvas.player_types[self.busy_body_id] == PlayerType.human_player:
            raise GinRummyProgramError("busy_body_id={} must be human player.".format(self.busy_body_id))
        raw_legal_actions = self.human_agent.state['raw_legal_actions']
        if self.game_canvas.query.is_scoring(legal_actions=raw_legal_actions):
            # 'boss' performs this, not human
            if not len(raw_legal_actions) == 1:
                raise GinRummyProgramError("len(legal_actions)={} must be 1.".format(len(raw_legal_actions)))
            action_id = raw_legal_actions[0]
            self._perform_score_action_id(action_id=action_id)
            return
        self._show_prolog_messages_on_human_turn()

    #
    #   Private methods to handle 'boss' play
    #
    #   The computer is the 'boss' who handles artifact moves injected into the move stream.
    #

    def _perform_deal_hand_move(self, move: DealHandMove):
        if utils.is_debug():
            print("{}".format(move))
        starting_new_game.show_new_game(game_canvas=self.game_canvas)
        self.env_thread.mark += 1
        self.busy_body_id = None

    def _perform_score_action_id(self, action_id: int):
        if utils.is_debug():
            if self.busy_body_id == 0:
                if not action_id == configurations.SCORE_PLAYER_0_ACTION_ID:
                    raise GinRummyProgramError("action_id={} must be SCORE_PLAYER_0_ACTION_ID".format(action_id))
            else:
                if not action_id == configurations.SCORE_PLAYER_1_ACTION_ID:
                    raise GinRummyProgramError("action_id={} must be SCORE_PLAYER_1_ACTION_ID".format(action_id))
        self.game_canvas.after_idle(self.did_perform_actions, [action_id])

    #
    #   Private methods to handle human play
    #

    def _show_prolog_messages_on_human_turn(self):
        legal_actions = self.human_agent.state['legal_actions']
        status_messaging.show_prolog_message(player_id=self.busy_body_id,
                                             legal_actions=legal_actions,
                                             game_canvas=self.game_canvas)

    #
    #   Private methods to handle computer play
    #

    def _show_prolog_messages_on_computer_turn(self):
        legal_actions = self.game_canvas.getter.get_legal_actions(player_id=self.busy_body_id)
        status_messaging.show_prolog_message(player_id=self.busy_body_id,
                                             legal_actions=legal_actions,
                                             game_canvas=self.game_canvas)

    def _advance_mark_for_computer_player(self):
        if not self.mark < len(self.moves):
            raise GinRummyProgramError("mark={} must be less than len(moves)={}.".format(self.mark, len(self.moves)))
        move = self.moves[self.mark]
        if isinstance(move, DealHandMove):
            self._perform_deal_hand_move(move=move)
        elif isinstance(move, DrawCardMove):
            self._perform_draw_card_move(move=move)
        elif isinstance(move, PickupDiscardMove):
            self._perform_pick_up_discard_move(move=move)
        elif isinstance(move, DeclareDeadHandMove):
            self._do_perform_declare_dead_hand_move(move=move)
        elif isinstance(move, DiscardMove):
            self._perform_discard_move(move=move)
        elif isinstance(move, KnockMove):
            self._perform_knock_move(move=move)
        elif isinstance(move, GinMove):
            self._perform_gin_move(move=move)
        elif isinstance(move, ScoreNorthMove) or isinstance(move, ScoreSouthMove):
            if utils.is_debug():
                print("{}".format(move))
            self._perform_score_action_id(action_id=move.action.action_id)

    def _perform_draw_card_move(self, move: DrawCardMove):
        if utils.is_debug():
            print("{}".format(move))
        player = move.player
        player_id = player.player_id
        card = move.card
        source_item_id = self.game_canvas.getter.get_top_stock_pile_item_id()
        source_card_item_id = self.game_canvas.getter.get_card_id(card_item_id=source_item_id)
        if not source_card_item_id == gin_rummy_utils.get_card_id(card=card):
            raise GinRummyProgramError("source_card_item_id={} doesn't match with card={}.".format(source_card_item_id, card))
        self.game_canvas.addtag_withtag(configurations.DRAWN_TAG, source_item_id)
        target_item_id = self.game_canvas.getter.get_held_pile_item_ids(player_id=player_id)[-1]
        target_item = self.game_canvas.canvas_item_by_item_id[target_item_id]
        handling_tap_held_pile.handle_tap_held_pile(hit_item=target_item, game_canvas=self.game_canvas)

    def _perform_pick_up_discard_move(self, move: PickupDiscardMove):
        if utils.is_debug():
            print("{}".format(move))
        player = move.player
        player_id = player.player_id
        card = move.card
        source_item_id = self.game_canvas.getter.get_top_discard_pile_item_id()
        source_card_item_id = self.game_canvas.getter.get_card_id(card_item_id=source_item_id)
        if not source_card_item_id == gin_rummy_utils.get_card_id(card=card):
            raise GinRummyProgramError("source_card_item_id={} doesn't match with card={}.".format(source_card_item_id, card))
        self.game_canvas.addtag_withtag(configurations.DRAWN_TAG, source_item_id)
        target_item_id = self.game_canvas.getter.get_held_pile_item_ids(player_id=player_id)[-1]
        target_item = self.game_canvas.canvas_item_by_item_id[target_item_id]
        handling_tap_held_pile.handle_tap_held_pile(hit_item=target_item, game_canvas=self.game_canvas)

    def _do_perform_declare_dead_hand_move(self, move: DeclareDeadHandMove):
        if utils.is_debug():
            print("{}".format(move))
        self.game_canvas.post_doing_action.post_do_declare_dead_hand_action(player_id=self.busy_body_id)
        self.game_canvas.after_idle(self.did_perform_actions, [move.action.action_id])

    def _perform_discard_move(self, move: DiscardMove):
        if utils.is_debug():
            print("{}".format(move))
        action_id = move.action.action_id
        if self.busy_body_id is None:
            raise GinRummyProgramError("busy_body_id cannot be None.")
        card_id = utils.get_action_card_id(action_id)
        source_item_id = self.game_canvas.card_item_ids[card_id]
        self.game_canvas.addtag_withtag(configurations.SELECTED_TAG, source_item_id)
        target_item_id = self.game_canvas.getter.get_top_discard_pile_item_id()
        if target_item_id is None:
            target_item_id = self.game_canvas.discard_pile_box_item
        if not self.game_canvas.is_treating_as_human(player_id=self.busy_body_id):
            # move source_item_id to end of held_pile invisibly
            self.game_canvas.tag_raise(source_item_id)
            utils.fan_held_pile(player_id=self.busy_body_id, game_canvas=self.game_canvas)
        handling_tap_discard_pile.handle_tap_discard_pile(hit_item=target_item_id, game_canvas=self.game_canvas)

    def _perform_knock_move(self, move: KnockMove):
        if utils.is_debug():
            print("{}".format(move))
        action_id = move.action.action_id
        card_id = utils.get_action_card_id(action_id)
        source_item_id = self.game_canvas.card_item_ids[card_id]
        self.game_canvas.addtag_withtag(configurations.SELECTED_TAG, source_item_id)
        self.game_canvas.post_doing_action.post_do_knock_action(source_item_id)

    def _perform_gin_move(self, move: GinMove):
        if utils.is_debug():
            print("{}".format(move))
        action_id = move.action.action_id
        card_id = utils.get_action_card_id(action_id)
        source_item_id = self.game_canvas.card_item_ids[card_id]
        self.game_canvas.addtag_withtag(configurations.SELECTED_TAG, source_item_id)
        self.game_canvas.post_doing_action.post_do_going_out_action(source_item_id)



'''
    File name: bridge/dealer.py
    Author: William Hale
    Date created: 11/25/2021
'''

from typing import List

from .player import BridgePlayer
from .utils.bridge_card import BridgeCard


class BridgeDealer:
    ''' Initialize a BridgeDealer dealer class
    '''
    def __init__(self, np_random):
        ''' set shuffled_deck, set stock_pile
        '''
        self.np_random = np_random
        self.shuffled_deck: List[BridgeCard] = BridgeCard.get_deck()  # keep a copy of the shuffled cards at start of new hand
        self.np_random.shuffle(self.shuffled_deck)
        self.stock_pile: List[BridgeCard] = self.shuffled_deck.copy()

    def deal_cards(self, player: BridgePlayer, num: int):
        ''' Deal some cards from stock_pile to one player

        Args:
            player (BridgePlayer): The BridgePlayer object
            num (int): The number of cards to be dealt
        '''
        for _ in range(num):
            player.hand.append(self.stock_pile.pop())



''' Limit Hold 'em rule model
'''
import rlcard
from rlcard.models.model import Model

class LimitholdemRuleAgentV1(object):
    ''' Limit Hold 'em Rule agent version 1
    '''

    def __init__(self):
        self.use_raw = True

    @staticmethod
    def step(state):
        ''' Predict the action when given raw state. A simple rule-based AI.
        Args:
            state (dict): Raw state from the game

        Returns:
            action (str): Predicted action
        '''
        legal_actions = state['raw_legal_actions']
        state = state['raw_obs']
        hand = state['hand']
        public_cards = state['public_cards']
        action = 'fold'
        # When having only 2 hand cards at the game start, choose fold to drop terrible cards:
        # Acceptable hand cards:
        # Pairs
        # AK, AQ, AJ, AT
        # A9s, A8s, ... A2s(s means flush)
        # KQ, KJ, QJ, JT
        # Fold all hand types except those mentioned above to save money
        if len(public_cards) == 0:
            if hand[0][1] == hand [1][1]:
                action = 'raise'
            elif hand[0][1] == 'A' or hand[1][1] == 'A':
                if 'K' in [hand[0][1], hand[1][1]] or 'Q' in [hand[0][1], hand[1][1]] or 'J' in [hand[0][1], hand[1][1]] or 'T' in [hand[0][1], hand[1][1]]:
                    action = 'raise'
                elif hand[0][0] == hand[1][0]:
                    action = 'raise'
            elif hand[0][1] == 'K' or hand[0][1] == 'Q' or hand[0][1] == 'J' or hand[0][1] == 'T':
                if hand[1][1] == 'K' or hand[1][1] == 'Q' or hand[1][1] == 'J' or hand[1][1] == 'T':
                    action = 'raise'
        if len(public_cards) == 3:
            public_cards_ranks = ['A', 'A', 'A']
            public_cards_flush = ['S', 'S', 'S']
            for i, _ in enumerate(public_cards):
                public_cards_ranks[i] = public_cards[i][1]
                public_cards_flush[i] = public_cards[i][0]
            if hand[0][1] == hand [1][1]:
            # if the player already have a pair, raise when public cards have card same as the pair
                if hand[0][1] in public_cards_ranks:
                    action = 'raise'
            elif hand[0][1] == 'A' or hand[1][1] == 'A':
                if 'K' in [hand[0][1], hand[1][1]] or 'Q' in [hand[0][1], hand[1][1]] or 'J' in [hand[0][1], hand[1][1]] or 'T' in [hand[0][1], hand[1][1]]:
                    # For AK, AQ, AJ, AT types, if public cards have A, K, Q, J, T, raise, because the chance of getting a straight greatly increases
                    if 'A' in public_cards_ranks or 'K' in public_cards_ranks or 'Q' in public_cards_ranks or 'J' in public_cards_ranks or 'T' in public_cards_ranks:
                        action = 'raise'
                # For A9s, A8s, ... A2s types, if public cards have same flush as the hand cards, raise, because the chance of getting a flush greatly increases
                elif hand[0][0] == hand[1][0]:
                    if hand[0][0] in public_cards_flush:
                        action = 'raise'
            elif max(public_cards_ranks) in ['5', '4' ,'3', '2']: # for KQ, KJ, QJ, JT, check when having no cards higher than 5
                action = 'check'
            else:
                action = 'call'

        if len(public_cards) == 5 or len(public_cards) == 4 :
            public_cards_ranks = []
            public_cards_flush = []
            for i, _ in enumerate(public_cards):
                public_cards_ranks.append('A')
                public_cards_flush.append('S')
                public_cards_ranks[i] = public_cards[i][1]
                public_cards_flush[i] = public_cards[i][0]
            if hand[0][1] == hand [1][1]:
            # if the player already have a pair, raise when public cards have card same as the pair
                if hand[0][1] in public_cards_ranks:
                    action = 'raise'
            elif hand[0][1] == 'A' or hand[1][1] == 'A':
                if 'K' in [hand[0][1], hand[1][1]] or 'Q' in [hand[0][1], hand[1][1]] or 'J' in [hand[0][1], hand[1][1]] or 'T' in [hand[0][1], hand[1][1]]:
                    # For AK, AQ, AJ, AT types, if public cards have A, K, Q, J, T, raise, because the chance of getting a straight greatly increases
                    if 'A' in public_cards_ranks or 'K' in public_cards_ranks or 'Q' in public_cards_ranks or 'J' in public_cards_ranks or 'T' in public_cards_ranks:
                        action = 'raise'
                    # For A9s, A8s, ... A2s types, if public cards have same flush as the hand cards, raise, because the chance of getting a flush greatly increases
                elif hand[0][0] == hand[1][0]:
                    if hand[0][0] in public_cards_flush:
                        action = 'raise'
            elif max(public_cards_ranks) in ['5', '4', '3', '2']: # for KQ, KJ, QJ, JT, fold when having no cards higher than 5
                action = 'fold'
            else:
                action = 'call'

        #return action
        if action in legal_actions:
            return action
        else:
            if action == 'raise':
                return 'call'
            if action == 'check':
                return 'fold'
            if action == 'call':
                return 'raise'
            else:
                return action

    def eval_step(self, state):
        ''' Step for evaluation. The same to step
        '''
        return self.step(state), []

class LimitholdemRuleModelV1(Model):
    ''' Limitholdem Rule Model version 1
    '''

    def __init__(self):
        ''' Load pretrained model
        '''
        env = rlcard.make('limit-holdem')

        rule_agent = LimitholdemRuleAgentV1()
        self.rule_agents = [rule_agent for _ in range(env.num_players)]

    @property
    def agents(self):
        ''' Get a list of agents for each position in a the game

        Returns:
            agents (list): A list of agents

        Note: Each agent should be just like RL agent with step and eval_step
              functioning well.
        '''
        return self.rule_agents

    @property
    def use_raw(self):
        ''' Indicate whether use raw state and action

        Returns:
            use_raw (boolean): True if using raw state and action
        '''
        return True



'''
    File name: bridge/utils/bridge_card.py
    Author: William Hale
    Date created: 11/25/2021
'''

from rlcard.games.base import Card


class BridgeCard(Card):

    suits = ['C', 'D', 'H', 'S']
    ranks = ['2', '3', '4', '5', '6', '7', '8', '9', 'T', 'J', 'Q', 'K', 'A']

    @staticmethod
    def card(card_id: int):
        return _deck[card_id]

    @staticmethod
    def get_deck() -> [Card]:
        return _deck.copy()

    def __init__(self, suit: str, rank: str):
        super().__init__(suit=suit, rank=rank)
        suit_index = BridgeCard.suits.index(self.suit)
        rank_index = BridgeCard.ranks.index(self.rank)
        self.card_id = 13 * suit_index + rank_index

    def __str__(self):
        return f'{self.rank}{self.suit}'

    def __repr__(self):
        return f'{self.rank}{self.suit}'


# deck is always in order from 2C, ... KC, AC, 2D, ... KD, AD, 2H, ... KH, AH, 2S, ... KS, AS
_deck = [BridgeCard(suit=suit, rank=rank) for suit in BridgeCard.suits for rank in BridgeCard.ranks]  # want this to be read-only



import unittest

import rlcard
from rlcard.agents.random_agent import RandomAgent
from .determism_util import is_deterministic


class TestDoudizhuEnv(unittest.TestCase):

    def test_reset_and_extract_state(self):
        env = rlcard.make('doudizhu')
        state, _ = env.reset()
        self.assertEqual(state['obs'].size, 790)

    def test_is_deterministic(self):
        self.assertTrue(is_deterministic('doudizhu'))

    def test_get_legal_actions(self):
        env = rlcard.make('doudizhu')
        env.set_agents([RandomAgent(env.num_actions) for _ in range(env.num_actions)])
        env.reset()
        legal_actions = env._get_legal_actions()
        for legal_action in legal_actions:
            self.assertLessEqual(legal_action, env.num_actions-1)

    def test_step(self):
        env = rlcard.make('doudizhu')
        _, player_id = env.reset()
        player = env.game.players[player_id]
        _, next_player_id = env.step(env.num_actions-2)
        self.assertEqual(next_player_id, (player.player_id+1)%len(env.game.players))

    def test_step_back(self):
        env = rlcard.make('doudizhu', config={'allow_step_back':True})
        _, player_id = env.reset()
        env.step(2)
        _, back_player_id = env.step_back()
        self.assertEqual(player_id, back_player_id)
        self.assertEqual(env.step_back(), False)

        env = rlcard.make('doudizhu')
        with self.assertRaises(Exception):
            env.step_back()

    def test_run(self):
        env = rlcard.make('doudizhu')
        env.set_agents([RandomAgent(env.num_actions) for _ in range(env.num_players)])
        trajectories, payoffs = env.run(is_training=False)
        self.assertEqual(len(trajectories), 3)
        win = []
        for player_id, payoff in enumerate(payoffs):
            if payoff == 1:
                win.append(player_id)
        if len(win) == 1:
            self.assertEqual(env.game.players[win[0]].role, 'landlord')
        if len(win) == 2:
            self.assertEqual(env.game.players[win[0]].role, 'peasant')
            self.assertEqual(env.game.players[win[1]].role, 'peasant')

    def test_decode_action(self):
        env = rlcard.make('doudizhu')
        env.reset()
        env.game.state['actions'] = ['33366', '33355']
        env.game.judger.playable_cards[0] = ['5', '6', '55', '555', '33366', '33355']
        decoded = env._decode_action(3)
        self.assertEqual(decoded, '6')
        env.game.state['actions'] = ['444', '44466', '44455']
        decoded = env._decode_action(29)
        self.assertEqual(decoded, '444')

    def test_get_perfect_information(self):
        env = rlcard.make('doudizhu')
        _, player_id = env.reset()
        self.assertEqual(player_id, env.get_perfect_information()['current_player'])
if __name__ == '__main__':
    unittest.main()



''' Doudizhu utils
'''
import os
import json
from collections import OrderedDict
import threading
import collections

import rlcard

# Read required docs
ROOT_PATH = rlcard.__path__[0]

if not os.path.isfile(os.path.join(ROOT_PATH, 'games/doudizhu/jsondata/action_space.txt')) \
        or not os.path.isfile(os.path.join(ROOT_PATH, 'games/doudizhu/jsondata/card_type.json')) \
        or not os.path.isfile(os.path.join(ROOT_PATH, 'games/doudizhu/jsondata/type_card.json')):
    import zipfile
    with zipfile.ZipFile(os.path.join(ROOT_PATH, 'games/doudizhu/jsondata.zip'),"r") as zip_ref:
        zip_ref.extractall(os.path.join(ROOT_PATH, 'games/doudizhu/'))

# Action space
action_space_path = os.path.join(ROOT_PATH, 'games/doudizhu/jsondata/action_space.txt')
with open(action_space_path, 'r') as f:
    ID_2_ACTION = f.readline().strip().split()
    ACTION_2_ID = {}
    for i, action in enumerate(ID_2_ACTION):
        ACTION_2_ID[action] = i

# a map of card to its type. Also return both dict and list to accelerate
card_type_path = os.path.join(ROOT_PATH, 'games/doudizhu/jsondata/card_type.json')
with open(card_type_path, 'r') as f:
    data = json.load(f, object_pairs_hook=OrderedDict)
    CARD_TYPE = (data, list(data), set(data))

# a map of type to its cards
type_card_path = os.path.join(ROOT_PATH, 'games/doudizhu/jsondata/type_card.json')
with open(type_card_path, 'r') as f:
    TYPE_CARD = json.load(f, object_pairs_hook=OrderedDict)

# rank list of solo character of cards
CARD_RANK_STR = ['3', '4', '5', '6', '7', '8', '9', 'T', 'J', 'Q', 'K',
                 'A', '2', 'B', 'R']
CARD_RANK_STR_INDEX = {'3': 0, '4': 1, '5': 2, '6': 3, '7': 4,
            '8': 5, '9': 6, 'T': 7, 'J': 8, 'Q': 9,
            'K': 10, 'A': 11, '2': 12, 'B': 13, 'R': 14}
# rank list
CARD_RANK = ['3', '4', '5', '6', '7', '8', '9', 'T', 'J', 'Q', 'K',
             'A', '2', 'BJ', 'RJ']

INDEX = {'3': 0, '4': 1, '5': 2, '6': 3, '7': 4,
         '8': 5, '9': 6, 'T': 7, 'J': 8, 'Q': 9,
         'K': 10, 'A': 11, '2': 12, 'B': 13, 'R': 14}
INDEX = OrderedDict(sorted(INDEX.items(), key=lambda t: t[1]))


def doudizhu_sort_str(card_1, card_2):
    ''' Compare the rank of two cards of str representation

    Args:
        card_1 (str): str representation of solo card
        card_2 (str): str representation of solo card

    Returns:
        int: 1(card_1 > card_2) / 0(card_1 = card2) / -1(card_1 < card_2)
    '''
    key_1 = CARD_RANK_STR.index(card_1)
    key_2 = CARD_RANK_STR.index(card_2)
    if key_1 > key_2:
        return 1
    if key_1 < key_2:
        return -1
    return 0


def doudizhu_sort_card(card_1, card_2):
    ''' Compare the rank of two cards of Card object

    Args:
        card_1 (object): object of Card
        card_2 (object): object of card
    '''
    key = []
    for card in [card_1, card_2]:
        if card.rank == '':
            key.append(CARD_RANK.index(card.suit))
        else:
            key.append(CARD_RANK.index(card.rank))
    if key[0] > key[1]:
        return 1
    if key[0] < key[1]:
        return -1
    return 0


def get_landlord_score(current_hand):
    ''' Roughly judge the quality of the hand, and provide a score as basis to
    bid landlord.

    Args:
        current_hand (str): string of cards. Eg: '56888TTQKKKAA222R'

    Returns:
        int: score
    '''
    score_map = {'A': 1, '2': 2, 'B': 3, 'R': 4}
    score = 0
    # rocket
    if current_hand[-2:] == 'BR':
        score += 8
        current_hand = current_hand[:-2]
    length = len(current_hand)
    i = 0
    while i < length:
        # bomb
        if i <= (length - 4) and current_hand[i] == current_hand[i+3]:
            score += 6
            i += 4
            continue
        # 2, Black Joker, Red Joker
        if current_hand[i] in score_map:
            score += score_map[current_hand[i]]
        i += 1
    return score

def cards2str_with_suit(cards):
    ''' Get the corresponding string representation of cards with suit

    Args:
        cards (list): list of Card objects

    Returns:
        string: string representation of cards
    '''
    return ' '.join([card.suit+card.rank for card in cards])

def cards2str(cards):
    ''' Get the corresponding string representation of cards

    Args:
        cards (list): list of Card objects

    Returns:
        string: string representation of cards
    '''
    response = ''
    for card in cards:
        if card.rank == '':
            response += card.suit[0]
        else:
            response += card.rank
    return response

class LocalObjs(threading.local):
    def __init__(self):
        self.cached_candidate_cards = None
_local_objs = LocalObjs()

def contains_cards(candidate, target):
    ''' Check if cards of candidate contains cards of target.

    Args:
        candidate (string): A string representing the cards of candidate
        target (string): A string representing the number of cards of target

    Returns:
        boolean
    '''
    # In normal cases, most continuous calls of this function
    #   will test different targets against the same candidate.
    # So the cached counts of each card in candidate can speed up
    #   the comparison for following tests if candidate keeps the same.
    if not _local_objs.cached_candidate_cards or _local_objs.cached_candidate_cards != candidate:
        _local_objs.cached_candidate_cards = candidate
        cards_dict = collections.defaultdict(int)
        for card in candidate:
            cards_dict[card] += 1
        _local_objs.cached_candidate_cards_dict = cards_dict
    cards_dict = _local_objs.cached_candidate_cards_dict
    if (target == ''):
        return True
    curr_card = target[0]
    curr_count = 1
    for card in target[1:]:
        if (card != curr_card):
            if (cards_dict[curr_card] < curr_count):
                return False
            curr_card = card
            curr_count = 1
        else:
            curr_count += 1
    if (cards_dict[curr_card] < curr_count):
        return False
    return True

def encode_cards(plane, cards):
    ''' Encode cards and represerve it into plane.

    Args:
        cards (list or str): list or str of cards, every entry is a
    character of solo representation of card
    '''
    if not cards:
        return None
    layer = 1
    if len(cards) == 1:
        rank = CARD_RANK_STR.index(cards[0])
        plane[layer][rank] = 1
        plane[0][rank] = 0
    else:
        for index, card in enumerate(cards):
            if index == 0:
                continue
            if card == cards[index-1]:
                layer += 1
            else:
                rank = CARD_RANK_STR.index(cards[index-1])
                plane[layer][rank] = 1
                layer = 1
                plane[0][rank] = 0
        rank = CARD_RANK_STR.index(cards[-1])
        plane[layer][rank] = 1
        plane[0][rank] = 0


def get_gt_cards(player, greater_player):
    ''' Provide player's cards which are greater than the ones played by
    previous player in one round

    Args:
        player (DoudizhuPlayer object): the player waiting to play cards
        greater_player (DoudizhuPlayer object): the player who played current biggest cards.

    Returns:
        list: list of string of greater cards

    Note:
        1. return value contains 'pass'
    '''
    # add 'pass' to legal actions
    gt_cards = ['pass']
    current_hand = cards2str(player.current_hand)
    target_cards = greater_player.played_cards
    target_types = CARD_TYPE[0][target_cards]
    type_dict = {}
    for card_type, weight in target_types:
        if card_type not in type_dict:
            type_dict[card_type] = weight
    if 'rocket' in type_dict:
        return gt_cards
    type_dict['rocket'] = -1
    if 'bomb' not in type_dict:
        type_dict['bomb'] = -1
    for card_type, weight in type_dict.items():
        candidate = TYPE_CARD[card_type]
        for can_weight, cards_list in candidate.items():
            if int(can_weight) > int(weight):
                for cards in cards_list:
                    # TODO: improve efficiency
                    if cards not in gt_cards and contains_cards(current_hand, cards):
                        # if self.contains_cards(current_hand, cards):
                        gt_cards.append(cards)
    return gt_cards



from rlcard.games.uno.dealer import UnoDealer as Dealer
from rlcard.games.uno.judger import UnoJudger as Judger
from rlcard.games.uno.player import UnoPlayer as Player
from rlcard.games.uno.round import UnoRound as Round
from rlcard.games.uno.game import UnoGame as Game




'''
    File name: bridge/utils/action_event.py
    Author: William Hale
    Date created: 11/25/2021
'''

from .bridge_card import BridgeCard

# ====================================
# Action_ids:
#       0 -> no_bid_action_id
#       1 to 35 -> bid_action_id (bid amount by suit or NT)
#       36 -> pass_action_id
#       37 -> dbl_action_id
#       38 -> rdbl_action_id
#       39 to 90 -> play_card_action_id
# ====================================


class ActionEvent(object):  # Interface

    no_bid_action_id = 0
    first_bid_action_id = 1
    pass_action_id = 36
    dbl_action_id = 37
    rdbl_action_id = 38
    first_play_card_action_id = 39

    def __init__(self, action_id: int):
        self.action_id = action_id

    def __eq__(self, other):
        result = False
        if isinstance(other, ActionEvent):
            result = self.action_id == other.action_id
        return result

    @staticmethod
    def from_action_id(action_id: int):
        if action_id == ActionEvent.pass_action_id:
            return PassAction()
        elif ActionEvent.first_bid_action_id <= action_id <= 35:
            bid_amount = 1 + (action_id - ActionEvent.first_bid_action_id) // 5
            bid_suit_id = (action_id - ActionEvent.first_bid_action_id) % 5
            bid_suit = BridgeCard.suits[bid_suit_id] if bid_suit_id < 4 else None
            return BidAction(bid_amount, bid_suit)
        elif action_id == ActionEvent.dbl_action_id:
            return DblAction()
        elif action_id == ActionEvent.rdbl_action_id:
            return RdblAction()
        elif ActionEvent.first_play_card_action_id <= action_id < ActionEvent.first_play_card_action_id + 52:
            card_id = action_id - ActionEvent.first_play_card_action_id
            card = BridgeCard.card(card_id=card_id)
            return PlayCardAction(card=card)
        else:
            raise Exception(f'ActionEvent from_action_id: invalid action_id={action_id}')

    @staticmethod
    def get_num_actions():
        ''' Return the number of possible actions in the game
        '''
        return 1 + 35 + 3 + 52  # no_bid, 35 bids, pass, dbl, rdl, 52 play_card


class CallActionEvent(ActionEvent):  # Interface
    pass


class PassAction(CallActionEvent):

    def __init__(self):
        super().__init__(action_id=ActionEvent.pass_action_id)

    def __str__(self):
        return "pass"

    def __repr__(self):
        return "pass"


class BidAction(CallActionEvent):

    def __init__(self, bid_amount: int, bid_suit: str or None):
        suits = BridgeCard.suits
        if bid_suit and bid_suit not in suits:
            raise Exception(f'BidAction has invalid suit: {bid_suit}')
        if bid_suit in suits:
            bid_suit_id = suits.index(bid_suit)
        else:
            bid_suit_id = 4
        bid_action_id = bid_suit_id + 5 * (bid_amount - 1) + ActionEvent.first_bid_action_id
        super().__init__(action_id=bid_action_id)
        self.bid_amount = bid_amount
        self.bid_suit = bid_suit

    def __str__(self):
        bid_suit = self.bid_suit
        if not bid_suit:
            bid_suit = 'NT'
        return f'{self.bid_amount}{bid_suit}'

    def __repr__(self):
        return self.__str__()


class DblAction(CallActionEvent):

    def __init__(self):
        super().__init__(action_id=ActionEvent.dbl_action_id)

    def __str__(self):
        return "dbl"

    def __repr__(self):
        return "dbl"


class RdblAction(CallActionEvent):

    def __init__(self):
        super().__init__(action_id=ActionEvent.rdbl_action_id)

    def __str__(self):
        return "rdbl"

    def __repr__(self):
        return "rdbl"


class PlayCardAction(ActionEvent):

    def __init__(self, card: BridgeCard):
        play_card_action_id = ActionEvent.first_play_card_action_id + card.card_id
        super().__init__(action_id=play_card_action_id)
        self.card: BridgeCard = card

    def __str__(self):
        return f"{self.card}"

    def __repr__(self):
        return f"{self.card}"
